# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A reader to read reviews and IMDb scores from
`Large Movie Reviews` dataset for sentiment classification.

There are two directories [pos/, neg/] for the reviews.
When using this reader, pass in 'data_dir/pos' and 'data_dir/neg',
respectively, as the file_path, to generate datapacks.

Dataset download link -
https://ai.stanford.edu/~amaas/data/sentiment/
Dataset Paper Citation -
@InProceedings{maas-EtAl:2011:ACL-HLT2011,
  author    = {Maas, Andrew L.  and  Daly, Raymond E.  and  Pham, Peter T.
                and  Huang, Dan  and  Ng, Andrew Y.  and  Potts, Christopher},
  title     = {Learning Word Vectors for Sentiment Analysis},
  booktitle = {Proceedings of the 49th Annual Meeting of the Association for
                    Computational Linguistics: Human Language Technologies},
  month     = {June},
  year      = {2011},
  address   = {Portland, Oregon, USA},
  publisher = {Association for Computational Linguistics},
  pages     = {142--150},
  url       = {http://www.aclweb.org/anthology/P11-1015}
}
"""

import os
import logging
import re
from typing import Iterator, List

from forte.data.data_pack import DataPack
from forte.data.data_utils_io import dataset_path_iterator
from forte.data.base_reader import PackReader
from ft.onto.base_ontology import Document, Sentence

__all__ = ["LargeMovieReader"]


class LargeMovieReader(PackReader):
    r""":class:`LargeMovieReader` is designed to read in
    the Large Movie Review Dataset v1.0.
    Reviews are stored in text files named
    following the convention [[id]_[rating].txt].
    """

    def __init__(self):
        super().__init__()
        self.REPLACE_NO_SPACE = re.compile(
            r"(\:)|(\')|(\,)|(\")|(\()|(\))|(\[)|(\])"
        )
        self.REPLACE_WITH_NEWLINE = re.compile(
            r"(<br\s*/><br\s*/>)|(\-)|(\/)|(\.)|(\;)|(\!)|(\?)"
        )

    def preprocess_reviews(self, para):
        para = self.REPLACE_NO_SPACE.sub("", para.lower())
        para = self.REPLACE_WITH_NEWLINE.sub("\n", para)
        return para

    def _collect(self, *args, **kwargs) -> Iterator[str]:
        # pylint: disable = unused-argument
        r"""Iterator over text files in the data_source.

        Args:
            args: args[0] is the directory to the pos/neg movie files.
            kwargs:

        Returns: Iterator over files in the path with txt extensions.
        """
        movie_directory: str = args[0]
        logging.info("Reading .txt from %s", movie_directory)
        return dataset_path_iterator(movie_directory, "txt")

    def _parse_pack(self, file_path: str) -> Iterator[DataPack]:
        data_pack: DataPack = DataPack()

        sent_begin: int = 0
        doc_text: str = ""

        with open(file_path, encoding="utf8") as doc:
            for para in doc:
                para = self.preprocess_reviews(para)
                sents = para.split("\n")
                for sent in sents:
                    if len(sent) > 0:
                        sent = sent.strip()
                        doc_text += sent + " "
                        doc_offset = sent_begin + len(sent) + 1
                        # Add sentences.
                        Sentence(data_pack, sent_begin, doc_offset - 1)
                        sent_begin = doc_offset

        pos_dir: str = os.path.basename(os.path.dirname(file_path))
        movie_file: str = os.path.basename(file_path)
        title: List = movie_file.split("_")
        doc_id: str = pos_dir + title[0]
        score: float = float(title[1].split(".")[0])
        score /= 10.0

        data_pack.pack_name = doc_id
        data_pack.set_text(doc_text)

        # Add documents.
        document: Document = Document(data_pack, 0, len(doc_text))
        document.sentiment = {doc_id: score}

        yield data_pack

    def _cache_key_function(self, movie_file: str) -> str:
        return os.path.basename(os.path.dirname(movie_file)) + os.path.basename(
            movie_file
        )



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Defines the basic data structures and interfaces for the Forte data
representation system.
"""
import uuid

from abc import abstractmethod, ABC
from collections.abc import MutableSequence, MutableMapping
from dataclasses import dataclass
from typing import (
    Iterable,
    Optional,
    Type,
    Hashable,
    TypeVar,
    Generic,
    Union,
    Dict,
    Iterator,
    overload,
    List,
)

import numpy as np

from packaging.version import Version

from forte.data.container import ContainerType, BasePointer

__all__ = [
    "Entry",
    "BaseLink",
    "BaseGroup",
    "LinkType",
    "GroupType",
    "EntryType",
    "Pointer",
    "MpPointer",
    "FDict",
    "FList",
    "FNdArray",
    "MultiEntry",
]

from forte.utils import get_full_module_name
from forte.version import DEFAULT_PACK_VERSION, PACK_ID_COMPATIBLE_VERSION

default_entry_fields = [
    "_Entry__pack",
    "_tid",
    "_embedding",
    "_span",
    "_begin",
    "_end",
    "_parent",
    "_child",
    "_members",
    "_Entry__field_modified",
    "field_records",
    "creation_records",
    "_id_manager",
]

unserializable_fields = [
    # This may be related to typing, but cannot be supported by typing.
    "__orig_class__",
]

_f_struct_keys: Dict[str, bool] = {}
_pointer_keys: Dict[str, bool] = {}


def set_state_func(instance, state):
    # pylint: disable=protected-access
    """
    An internal used function. `instance` is an instance of Entry or a
    MultiEntry. This function will populate the internal states for them.

    Args:
        instance:
        state:

    Returns:
        None
    """
    # During de-serialization, convert the list back to numpy array.
    if "_embedding" in state:
        state["_embedding"] = np.array(state["_embedding"])
    else:
        state["_embedding"] = np.empty(0)

    # NOTE: the __pack will be set via set_pack from the Pack side.
    cls_name = get_full_module_name(instance)
    for k, v in state.items():
        key = cls_name + "_" + k
        if _f_struct_keys.get(key, False):
            v._set_parent(instance)
        else:
            if isinstance(v, (FList, FDict)):
                v._set_parent(instance)
                _f_struct_keys[key] = True
            else:
                _f_struct_keys[key] = False

    instance.__dict__.update(state)


def get_state_func(instance):
    # pylint: disable=protected-access
    r"""In serialization, the reference to pack is not set, and
    it will be set by the container.

    This also implies that it is not advised to serialize an entry on its
    own, without the ``Container`` as the context, there is little semantics
    remained in an entry and unexpected errors could occur.
    """
    state = instance.__dict__.copy()
    # During serialization, convert the numpy array as a list.
    emb = list(instance._embedding.tolist())
    if len(emb) == 0:
        state.pop("_embedding")
    else:
        state["_embedding"] = emb

    cls_name = get_full_module_name(instance)
    for k, v in state.items():
        key = cls_name + "_" + k
        if k in _pointer_keys:
            if _pointer_keys[key]:
                state[k] = v.as_pointer(instance)
        else:
            if isinstance(v, Entry):
                state[k] = v.as_pointer(instance)
                _pointer_keys[key] = True
            else:
                _pointer_keys[key] = False

    state.pop("_Entry__pack")
    return state


@dataclass
class Entry(Generic[ContainerType]):
    r"""The base class inherited by all NLP entries. This is the main data type
    for all in-text NLP analysis results. The main sub-types are
    :class:`~forte.data.ontology.top.Annotation`, ``Link``, ``Generics``, and
    ``Group``.

    An :class:`forte.data.ontology.top.Annotation` object represents a
    span in text.

    A :class:`forte.data.ontology.top.Link` object represents a binary
    link relation between two entries.

    A :class:`forte.data.ontology.top.Generics` object.

    A :class:`forte.data.ontology.top.Group` object represents a
    collection of multiple entries.

    Main Attributes:

        - embedding: The embedding vectors (numpy array of floats) of this
          entry.

    Args:
        pack: Each entry should be associated with one pack upon creation.
    """

    def __init__(self, pack: ContainerType):
        # The Entry should have a reference to the data pack, and the data pack
        # need to store the entries. In order to resolve the cyclic references,
        # we create a generic class EntryContainer to be the place holder of
        # the actual. Whether this entry can be added to the pack is delegated
        # to be checked by the pack.
        super().__init__()
        self.__pack: ContainerType = pack
        self._tid: int = uuid.uuid4().int
        self._embedding: np.ndarray = np.empty(0)
        self.pack._validate(self)
        self.pack.on_entry_creation(self)

    def regret_creation(self):
        self.__pack.regret_creation(self)

    def __getstate__(self):
        r"""In serialization, the pack is not serialize, and it will be set
        by the container.

        This also implies that it is not advised to serialize an entry on its
        own, without the ``Container`` as the context, there is little semantics
        remained in an entry.
        """
        return get_state_func(self)

    def __setstate__(self, state):
        set_state_func(self, state)

    # using property decorator
    # a getter function for self._embedding
    @property
    def embedding(self):
        r"""Get the embedding vectors (numpy array of floats) of the entry."""
        return self._embedding

    # a setter function for self._embedding
    @embedding.setter
    def embedding(self, embed):
        r"""Set the embedding vectors of the entry.

        Args:
            embed: The embedding vectors which can be numpy array of floats or
                list of floats.
        """
        self._embedding = np.array(embed)

    @property
    def tid(self) -> int:
        """
        Get the id of this entry.

        Returns:
            id of the entry
        """
        return self._tid

    @property
    def pack(self) -> ContainerType:
        return self.__pack

    @property
    def pack_id(self) -> int:
        """
        Get the id of the pack that contains this entry.

        Returns:
            id of the pack that contains this entry.
        """
        return self.__pack.pack_id  # type: ignore

    def set_pack(self, pack: ContainerType):
        self.__pack = pack

    def relink_pointer(self):
        """
        This function is normally called after deserialization. It can be called
        when the pack reference of this entry is ready (i.e. after `set_pack`).
        The purpose is to convert the `Pointer` objects into actual entries.
        """
        cls_name = get_full_module_name(self)
        for k, v in self.__dict__.items():
            key = cls_name + "_" + k
            if k in _pointer_keys:
                if _pointer_keys[key]:
                    setattr(self, k, self._resolve_pointer(v))
            else:
                if isinstance(v, BasePointer):
                    _pointer_keys[key] = True
                    setattr(self, k, self._resolve_pointer(v))
                else:
                    _pointer_keys[key] = False

    def as_pointer(self, from_entry: "Entry"):
        """
        Return this entry as a pointer of this entry relative to the
        ``from_entry``.

        Args:
            from_entry: the entry to point from.

        Returns:
            A pointer to the this entry from the ``from_entry``.
        """
        if isinstance(from_entry, MultiEntry):
            return MpPointer(
                # bug fix/enhancement 559: change pack index to pack_id for multi-entry/multi-pack
                self.pack_id,
                self.tid,  # from_entry.pack.get_pack_index(self.pack_id)
            )
        elif isinstance(from_entry, Entry):
            return Pointer(self.tid)

    def _resolve_pointer(self, ptr: BasePointer):
        """
        Resolve into an entry on the provided pointer ``ptr`` from this entry.

        Args:
            ptr: A pointer that refer to an entity.

        Returns:
            None
        """
        if isinstance(ptr, Pointer):
            return self.pack.get_entry(ptr.tid)
        else:
            raise TypeError(
                f"Unsupported pointer type {ptr.__class__} for entry"
            )

    def entry_type(self) -> str:
        """Return the full name of this entry type."""
        module = self.__class__.__module__
        if module is None or module == str.__class__.__module__:
            return self.__class__.__name__
        else:
            return module + "." + self.__class__.__name__

    # def _check_attr_type(self, key, value):
    #     """
    #     Use the type hint to validate whether the provided value is as expected.
    #
    #     Args:
    #         key:  The field name.
    #         value: The field value.
    #
    #     Returns:
    #
    #     """
    #     if key not in default_entry_fields:
    #         hints = get_type_hints(self.__class__)
    #         if key not in hints.keys():
    #             warnings.warn(
    #                 f"Base on attributes in entry definition, "
    #                 f"the [{key}] attribute_name does not exist in the "
    #                 f"[{type(self).__name__}] that you specified to add to."
    #             )
    #         is_valid = check_type(value, hints[key])
    #         if not is_valid:
    #             warnings.warn(
    #                 f"Based on type annotation, "
    #                 f"the [{key}] attribute of [{type(self).__name__}] "
    #                 f"should be [{hints[key]}], but got [{type(value)}]."
    #             )

    # def __setattr__(self, key, value):
    #     super().__setattr__(key, value)
    #
    #     if isinstance(value, Entry):
    #         if value.pack == self.pack:
    #             # Save a pointer to the value from this entry.
    #             self.__dict__[key] = Pointer(value.tid)
    #         else:
    #             raise PackDataException(
    #                 "An entry cannot refer to entries in another data pack."
    #             )
    #     else:
    #         super().__setattr__(key, value)
    #
    #     # We add the record to the system.
    #     if key not in default_entry_fields:
    #         self.__pack.record_field(self.tid, key)

    # def __getattribute__(self, item):
    #     try:
    #         v = super().__getattribute__(item)
    #     except AttributeError:
    #         # For all unknown attributes, return None.
    #         return None
    #
    #     if isinstance(v, BasePointer):
    #         # Using the pointer to get the entry.
    #         return self._resolve_pointer(v)
    #     else:
    #         return v

    def __eq__(self, other):
        r"""
        The eq function for :class:`~forte.data.ontology.core.Entry` objects.
        Can be further implemented in each subclass.

        Args:
            other:

        Returns:
            None
        """
        if other is None:
            return False

        return (type(self), self._tid) == (type(other), other.tid)

    def __lt__(self, other):
        r"""By default, compared based on type string."""
        return (str(type(self))) < (str(type(other)))

    def __hash__(self) -> int:
        r"""The hash function for :class:`~forte.data.ontology.core.Entry` objects.
        To be implemented in each subclass.
        """
        return hash((type(self), self._tid))

    @property
    def index_key(self) -> Hashable:
        # Think about how to use the index key carefully.
        return self._tid


class MultiEntry(Entry, ABC):
    r"""The base class for multi-pack entries. The main sub-types are
    ``MultiPackLink``, ``MultiPackGenerics``, and ``MultiPackGroup``.

    A :class:`forte.data.ontology.top.MultiPackLink` object represents a binary
    link relation between two entries between different data packs.

    A :class:`forte.data.ontology.top.MultiPackGroup` object represents a
    collection of multiple entries among different data packs.
    """

    def as_pointer(self, from_entry: "Entry") -> "Pointer":
        """
        Get a pointer of the entry relative to this entry

        Args:
            from_entry: The entry relative from.

        Returns:
             A pointer relative to the this entry.
        """
        if isinstance(from_entry, MultiEntry):
            return Pointer(self.tid)
        elif isinstance(from_entry, Entry):
            raise ValueError(
                "Do not support reference a multi pack entry from an entry."
            )

    def _resolve_pointer(self, ptr: BasePointer) -> Entry:
        if isinstance(ptr, Pointer):
            return self.pack.get_entry(ptr.tid)
        elif isinstance(ptr, MpPointer):
            # bugfix/new feature 559: in new version pack_index will be using pack_id internally
            pack_array_index = ptr.pack_index  # old version
            pack_version = ""
            try:
                pack_version = self.pack.pack_version
            except AttributeError:
                pack_version = DEFAULT_PACK_VERSION  # set to default if lacking version attribute

            if Version(pack_version) >= Version(PACK_ID_COMPATIBLE_VERSION):
                pack_array_index = self.pack.get_pack_index(
                    ptr.pack_index
                )  # default: new version

            return self.pack.packs[pack_array_index].get_entry(ptr.tid)
        else:
            raise TypeError(f"Unknown pointer type {ptr.__class__}")


EntryType = TypeVar("EntryType", bound=Entry)

ParentEntryType = TypeVar("ParentEntryType", bound=Entry)


class FList(Generic[ParentEntryType], MutableSequence):
    """
    FList allows the elements to be Forte entries. FList will internally
    stores the entry as their tid to avoid nesting.
    """

    def __init__(
        self,
        parent_entry: ParentEntryType,
        data: Optional[Iterable[EntryType]] = None,
    ):
        super().__init__()
        self.__parent_entry = parent_entry
        self.__data: List[BasePointer] = []
        if data is not None:
            self.__data = [d.as_pointer(self.__parent_entry) for d in data]

    def __eq__(self, other):
        return self.__data == other._FList__data

    def __getstate__(self):
        state = self.__dict__.copy()
        # Parent entry cannot be serialized, should be set by the parent with
        #  _set_parent.
        state.pop("_FList__parent_entry")

        state["data"] = state.pop("_FList__data")

        # We make a copy of the whole state but there are items cannot be
        # serialized.
        for f in unserializable_fields:
            if f in state:
                state.pop(f)
        return state

    def __setstate__(self, state):
        state["_FList__data"] = state.pop("data")
        self.__dict__.update(state)

    def _set_parent(self, parent_entry: ParentEntryType):
        self.__parent_entry = parent_entry

    def insert(self, index: int, entry: EntryType):
        self.__data.insert(index, entry.as_pointer(self.__parent_entry))

    @overload
    @abstractmethod
    def __getitem__(self, i: int) -> EntryType:
        ...

    @overload
    @abstractmethod
    def __getitem__(self, s: slice) -> MutableSequence:
        ...

    def __getitem__(
        self, index: Union[int, slice]
    ) -> Union[EntryType, MutableSequence]:
        if isinstance(index, slice):
            return [
                self.__parent_entry._resolve_pointer(d)
                for d in self.__data[index]
            ]
        else:
            return self.__parent_entry._resolve_pointer(self.__data[index])

    def __setitem__(
        self,
        index: Union[int, slice],
        value: Union[EntryType, Iterable[EntryType]],
    ) -> None:
        if isinstance(index, int):
            self.__data[index] = value.as_pointer(  # type: ignore
                self.__parent_entry
            )
        else:
            self.__data[index] = [
                v.as_pointer(self.__parent_entry) for v in value  # type: ignore
            ]

    def __delitem__(self, index: Union[int, slice]) -> None:
        del self.__data[index]

    def __len__(self) -> int:
        return len(self.__data)


KeyType = TypeVar("KeyType", bound=Hashable)
ValueType = TypeVar("ValueType", bound=Entry)


class FDict(Generic[KeyType, ValueType], MutableMapping):
    """
    FDict allows the values to be Forte entries. FDict will internally
    stores the entry as their tid to avoid nesting. Note that key is not
    supported to be entries now.
    """

    def __init__(
        self,
        parent_entry: ParentEntryType,
        data: Optional[Dict[KeyType, ValueType]] = None,
    ):
        super().__init__()

        self.__parent_entry = parent_entry
        self.__data: Dict[KeyType, BasePointer] = {}

        if data is not None:
            self.__data = {
                k: v.as_pointer(self.__parent_entry) for k, v in data.items()
            }

    def _set_parent(self, parent_entry: ParentEntryType):
        self.__parent_entry = parent_entry

    def __eq__(self, other):
        return self.__data == other._FDict__data

    def __getstate__(self):
        state = self.__dict__.copy()
        # The __parent_entry need to be assigned via its parent entry,
        # so a serialized dict may not have the following key ready sometimes.
        state.pop("_FDict__parent_entry")

        state["data"] = state.pop("_FDict__data")

        # We make a copy of the whole state but there are items cannot be
        # serialized.
        for f in unserializable_fields:
            if f in state:
                state.pop(f)
        return state

    def __setstate__(self, state):
        state["_FDict__data"] = state.pop("data")
        self.__dict__.update(state)

    def __setitem__(self, k: KeyType, v: ValueType) -> None:
        try:
            self.__data[k] = v.as_pointer(self.__parent_entry)
        except AttributeError as e:
            raise AttributeError(
                f"Item of the FDict must be of type entry, "
                f"got {v.__class__}"
            ) from e

    def __delitem__(self, k: KeyType) -> None:
        del self.__data[k]

    def __getitem__(self, k: KeyType) -> ValueType:
        return self.__parent_entry._resolve_pointer(self.__data[k])

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self) -> Iterator[KeyType]:
        yield from self.__data


class FNdArray:
    """
    FNdArray is a wrapper of a NumPy array that stores shape and data type
    of the array if they are specified. Only when both shape and data type
    are provided, will FNdArray initialize a placeholder array through
    np.ndarray(shape, dtype=dtype).
    More details about np.ndarray(...):
    https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html
    """

    def __init__(
        self, dtype: Optional[str] = None, shape: Optional[List[int]] = None
    ):
        super().__init__()
        self._dtype: Optional[np.dtype] = (
            np.dtype(dtype) if dtype is not None else dtype
        )
        self._shape: Optional[tuple] = (
            tuple(shape) if shape is not None else shape
        )
        self._data: Optional[np.ndarray] = None
        if dtype and shape:
            self._data = np.ndarray(shape, dtype=dtype)

    @property
    def dtype(self):
        return self._dtype

    @property
    def shape(self):
        return self._shape

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, array: Union[np.ndarray, List]):
        if isinstance(array, np.ndarray):
            if self.dtype and not np.issubdtype(array.dtype, self.dtype):
                raise TypeError(
                    f"Expecting type or subtype of {self.dtype}, but got {array.dtype}."
                )
            if self.shape and self.shape != array.shape:
                raise AttributeError(
                    f"Expecting shape {self.shape}, but got {array.shape}."
                )
            self._data = array

        elif isinstance(array, list):
            array_np = np.array(array, dtype=self.dtype)
            if self.shape and self.shape != array_np.shape:
                raise AttributeError(
                    f"Expecting shape {self.shape}, but got {array_np.shape}."
                )
            self._data = array_np

        else:
            raise ValueError(
                f"Can only accept numpy array or python list, but got {type(array)}"
            )

        # Stored dtype and shape should match to the provided array's.
        self._dtype = self._data.dtype
        self._shape = self._data.shape


class Pointer(BasePointer):
    """
    A pointer that points to an entry in the current pack, this is basically
    containing the entry's tid.
    """

    def __init__(self, tid: int):
        self._tid: int = tid

    @property
    def tid(self):
        return self._tid

    def __str__(self):
        return f"[Entry Pointer]:{self.tid}"

    def __eq__(self, other):
        return self.tid == other.tid


class MpPointer(BasePointer):
    """
    Multi pack Pointer. A pointer that refers to an entry of one of the pack in
    the multi pack. This contains the pack's index and the entries' tid.
    """

    def __init__(self, pack_index: int, tid: int):
        self._pack_index: int = pack_index
        self._tid: int = tid

    @property
    def pack_index(self):
        return self._pack_index

    @property
    def tid(self):
        return self._tid

    def __str__(self):
        return f"[Entry Pointer]:{self.pack_index},{self.tid}"

    def __eq__(self, other):
        return self._pack_index == other._pack_index and self.tid == other.tid


class BaseLink(Entry, ABC):
    def __init__(
        self,
        pack: ContainerType,
        parent: Optional[Entry] = None,
        child: Optional[Entry] = None,
    ):
        super().__init__(pack)

        if parent is not None:
            self.set_parent(parent)
        if child is not None:
            self.set_child(child)

    @abstractmethod
    def set_parent(self, parent: Entry):
        r"""This will set the `parent` of the current instance with given Entry
        The parent is saved internally by its pack specific index key.

        Args:
            parent: The parent entry.
        """
        raise NotImplementedError

    @abstractmethod
    def set_child(self, child: Entry):
        r"""This will set the `child` of the current instance with given Entry
        The child is saved internally by its pack specific index key.

        Args:
            child: The child entry
        """
        raise NotImplementedError

    @abstractmethod
    def get_parent(self) -> Entry:
        r"""Get the parent entry of the link.

        Returns:
             An instance of :class:`~forte.data.ontology.core.Entry` that is the child of the link
             from the given :class:`~forte.data.data_pack.DataPack`.
        """
        raise NotImplementedError

    @abstractmethod
    def get_child(self) -> Entry:
        r"""Get the child entry of the link.

        Returns:
             An instance of :class:`~forte.data.ontology.core.Entry` that is the child of the link
             from the given :class:`~forte.data.data_pack.DataPack`.
        """
        raise NotImplementedError

    def __eq__(self, other):
        if other is None:
            return False
        return (type(self), self.get_parent(), self.get_child()) == (
            type(other),
            other.get_parent(),
            other.get_child(),
        )

    def __hash__(self):
        return hash((type(self), self.get_parent(), self.get_child()))

    @property
    def index_key(self) -> int:
        return self.tid


class BaseGroup(Entry, Generic[EntryType]):
    r"""Group is an entry that represent a group of other entries. For example,
    a "coreference group" is a group of coreferential entities. Each group will
    store a set of members, no duplications allowed.

    This is the :class:`BaseGroup` interface. Specific member constraints are
    defined in the inherited classes.
    """
    MemberType: Type[EntryType]

    def __init__(
        self, pack: ContainerType, members: Optional[Iterable[EntryType]] = None
    ):
        super().__init__(pack)
        if members is not None:
            self.add_members(members)

    @abstractmethod
    def add_member(self, member: EntryType):
        r"""Add one entry to the group.

        Args:
            member: One member to be added to the group.
        """
        raise NotImplementedError

    def add_members(self, members: Iterable[EntryType]):
        r"""Add members to the group.

        Args:
            members: An iterator of members to be added to the group.
        """
        for member in members:
            self.add_member(member)

    def __hash__(self):
        r"""The hash function of :class:`Group`.

        Users can define their own hash function by themselves but this must
        be consistent to :meth:`eq`.
        """
        return hash((type(self), tuple(self.get_members())))

    def __eq__(self, other):
        r"""The eq function of :class:`Group`. By default, :class:`Group`
        objects are regarded as the same if they have the same type, members,
        and are generated by the same component.

        Users can define their own eq function by themselves but this must
        be consistent to :meth:`hash`.
        """
        if other is None:
            return False
        return (type(self), self.get_members()) == (
            type(other),
            other.get_members(),
        )

    @abstractmethod
    def get_members(self) -> List[EntryType]:
        r"""Get the member entries in the group.

        Returns:
             Instances of :class:`~forte.data.ontology.core.Entry` that are the members of the
             group.
        """
        raise NotImplementedError

    @property
    def index_key(self) -> int:
        return self.tid


GroupType = TypeVar("GroupType", bound=BaseGroup)
LinkType = TypeVar("LinkType", bound=BaseLink)



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Here we define some utility functions for CoNLL evaluation datasets change.
We can add other datasets conversion function for CoNLL here in the future.
"""
from typing import List, Dict, Optional, Type, Tuple, Union, Callable

from forte.data.data_pack import DataPack
from forte.data.ontology import Annotation
from ft.onto.base_ontology import Sentence


def post_edit(element: Tuple[Optional[str], str]) -> str:
    r"""Post process the tags. Convert it from tuple
    to string.

    Args:
        element:
            a tuple of BIO tag with element[0]
            as tag label and element[1] as
            BIO, e.g. `("PER", "B")`.

    Returns:
        BIO tag in string format, e.g. `"B-PER"`.

    """
    if element[0] is None:
        return "O"
    return "%s-%s" % (element[1], element[0])


def bio_tagging(
    pack: DataPack,
    tagging_unit_type: Type[Annotation],
    entry_type: Type[Annotation],
    attribute: Union[Callable[[Annotation], str], str],
    context: Optional[Annotation] = None,
) -> List[Tuple[Optional[str], str]]:
    """This utility function use BIO tagging method to convert tags
    of "instance_entry" into the same length as "instance_tagging_unit". Both
    element from "instance_entry" and "instance_tagging_unit" should Annotation
    type. This function uses their position information to
    determine B, I, O tagging for the entry on each tagging_unit element.

    Args:
        pack: The datapack that contains the current
            instance.

        tagging_unit_type: The type of tagging unit that entry
            tag should align to. For example, it can be Token, which means
            returned tags should aligned to tokens in one sentence.

        entry_type: The type of entry that contains tags. For
            example, it can be EntityMention, which means tags comes from the
            EntityMention of one sentence. Note that the number of EntityMention
            can be different from the number of Token. That is why we need to
            use BIO tagging to align them.

        attribute: A function to
            get the tags via the attribute of an entry. Or a str of the name
            of the attribute. For example, it can be "ner_type", which means
            the attribute ner_type of the entry will be treated as tags.

        context: The instance from which the
            extractor will extractor feature. For example, an instance of
            Sentence type, which mean the tagging sequence comes from
            one sentence. If None, then the whole data pack will be used.

    Returns:
        A list of the type List[Tuple[Optional[str], str]]. For example,
        [(None, "O"), (LOC, "B"), (LOC, "I"), (None, "O"),
         (None, "O"), (PER, "B"), (None, "O")]
    """

    # Tokens in the sentence.
    tagging_units: List[Annotation] = list(pack.get(tagging_unit_type, context))
    # All mentions in the sentence.
    instance_entry: List[Annotation] = list(pack.get(entry_type, context))

    tagged: List[Tuple[Optional[str], str]] = []
    unit_id = 0
    entry_id = 0
    while unit_id < len(tagging_units) or entry_id < len(instance_entry):
        if entry_id == len(instance_entry):
            tagged.append((None, "O"))
            unit_id += 1
            continue

        is_start = True
        for unit in pack.get(tagging_unit_type, instance_entry[entry_id]):
            while (
                unit_id < len(tagging_units)
                and tagging_units[unit_id].begin != unit.begin
                and tagging_units[unit_id].end != unit.end
            ):
                tagged.append((None, "O"))
                unit_id += 1

            if is_start:
                location = "B"
                is_start = False
            else:
                location = "I"

            unit_id += 1
            if callable(attribute):
                tagged.append((attribute(instance_entry[entry_id]), location))
            else:
                tagged.append(
                    (getattr(instance_entry[entry_id], attribute), location)
                )
        entry_id += 1
    return tagged


def get_tag(
    pack: DataPack,
    instance: Annotation,
    tagging_unit: Type[Annotation],
    entry_type: Type[Annotation],
    attribute: str,
) -> List[str]:
    r"""Align entries to tagging units, and convert it to string format.

    Args:
        pack: The datapack that contains the current
            instance.

        instance: The instance from which the
            extractor will extractor feature. For example, an instance of
            Sentence type, which mean the tagging sequence comes from
            one sentence.

        tagging_unit: The type of tagging unit that entry
            tag should align to. For example, it can be Token, which means
            returned tags should aligned to tokens in one sentence.

        entry_type: The type of entry that contains tags. For
            example, it can be EntityMethion, which means tags comes from the
            EntityMention of one sentence. Note that the number of EntityMention
            can be different from the number of Token. That is why we need to
            use BIO tagging to aglin them.

        attribute: A str of the name of the attribute. For example,
            it can be "ner_type", which means the attribute ner_type of
            the entry will be treated as tags.

    Returns:
        BIO tag sequence in string format.

    """
    tag = bio_tagging(pack, tagging_unit, entry_type, attribute, instance)
    tag = [post_edit(x) for x in tag]
    return tag


def write_tokens_to_file(
    pred_pack: DataPack,
    refer_pack: DataPack,
    refer_request: Dict,
    tagging_unit: Type[Annotation],
    entry_type: Type[Annotation],
    attribute: str,
    output_file: str,
):
    r"""Write prediction results into files, along with reference
    labels, for performance evaluation.

    Args:
        pred_pack: The predicated datapack.

        refer_pack: The reference datapack.

        refer_request: Reference request.

        tagging_unit: The type of tagging unit that entry
            tag should align to. For example, it can be Token, which means
            returned tags should aligned to tokens in one sentence.

        entry_type: The type of entry that contains tags. For
            example, it can be EntityMethion, which means tags comes from the
            EntityMention of one sentence. Note that the number of EntityMention
            can be different from the number of Token. That is why we need to
            use BIO tagging to aglin them.

        attribute: A str of the name of the attribute. For example,
            it can be "ner_type", which means the attribute ner_type of the
            entry will be treated as tags.

        output_file: The path where results write.

    """
    with open(output_file, "a+", encoding="utf-8") as opened_file:
        for refer_data, pred_sent, refer_sent in zip(
            refer_pack.get_data(**refer_request),
            pred_pack.get(Sentence),
            refer_pack.get(Sentence),
        ):

            refer_tag = get_tag(
                refer_pack, refer_sent, tagging_unit, entry_type, attribute
            )
            pred_tag = get_tag(
                pred_pack, pred_sent, tagging_unit, entry_type, attribute
            )

            words = refer_data["Token"]["text"]

            for i, (word, tgt, pred) in enumerate(
                zip(words, refer_tag, pred_tag), 1
            ):
                opened_file.write("%d %s %s %s\n" % (i, word, tgt, pred))
            opened_file.write("\n")



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Dict, List, Tuple, Optional
from forte.utils import create_import_error_msg
from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.data_pack import DataPack
from forte.data.ontology import Annotation
from forte.data.span import Span
from forte.models.srl.model import LabeledSpanGraphNetwork
from forte.processors.base.batch_processor import RequestPackingProcessor
from ft.onto.base_ontology import (
    PredicateLink,
    PredicateMention,
    PredicateArgument,
)

try:
    import texar.torch as tx
except ImportError as e:
    raise ImportError(
        create_import_error_msg("texar-pytorch", "nlp", "Texar models")
    ) from e


try:
    import torch
except ImportError as e:
    raise ImportError(
        create_import_error_msg("torch", "nlp", "nlp processors")
    ) from e


logger = logging.getLogger(__name__)

__all__ = [
    "SRLPredictor",
]

Prediction = List[Tuple[Span, List[Tuple[Span, str]]]]


class SRLPredictor(RequestPackingProcessor):
    """
    An Semantic Role labeler trained according to `He, Luheng, et al.
    "Jointly predicting predicates and arguments in neural semantic role
    labeling." <https://aclweb.org/anthology/P18-2058>`_.
    """

    word_vocab: tx.data.Vocab
    char_vocab: tx.data.Vocab
    model: LabeledSpanGraphNetwork

    def __init__(self):
        super().__init__()
        self.device = torch.device(
            torch.cuda.current_device() if torch.cuda.is_available() else "cpu"
        )

    def initialize(self, resources: Resources, configs: Optional[Config]):
        super().initialize(resources, configs)

        model_dir = configs.storage_path if configs is not None else None
        logger.info("restoring SRL model from %s", model_dir)

        # initialize the batcher
        if configs:
            self.batcher.initialize(configs.batcher)

        self.word_vocab = tx.data.Vocab(
            os.path.join(model_dir, "embeddings/word_vocab.english.txt")
        )
        self.char_vocab = tx.data.Vocab(
            os.path.join(model_dir, "embeddings/char_vocab.english.txt")
        )
        model_hparams = LabeledSpanGraphNetwork.default_hparams()
        model_hparams["context_embeddings"]["path"] = os.path.join(
            model_dir, model_hparams["context_embeddings"]["path"]
        )
        model_hparams["head_embeddings"]["path"] = os.path.join(
            model_dir, model_hparams["head_embeddings"]["path"]
        )
        self.model = LabeledSpanGraphNetwork(
            self.word_vocab, self.char_vocab, model_hparams
        )
        self.model.load_state_dict(
            torch.load(
                os.path.join(model_dir, "pretrained/model.pt"),
                map_location=self.device,
            )
        )
        self.model.eval()

    def predict(self, data_batch: Dict) -> Dict[str, List[Prediction]]:
        text: List[List[str]] = [
            sentence.tolist() for sentence in data_batch["Token"]["text"]
        ]
        text_ids, length = tx.data.padded_batch(
            [
                self.word_vocab.map_tokens_to_ids_py(sentence)
                for sentence in text
            ]
        )
        text_ids = torch.from_numpy(text_ids).to(device=self.device)
        length = torch.tensor(length, dtype=torch.long, device=self.device)
        batch_size = len(text)
        batch = tx.data.Batch(
            batch_size,
            text=text,
            text_ids=text_ids,
            length=length,
            srl=[[]] * batch_size,
        )
        self.model = self.model.to(self.device)
        batch_srl_spans = self.model.decode(batch)

        # Convert predictions into annotations.
        batch_predictions: List[Prediction] = []
        for idx, srl_spans in enumerate(batch_srl_spans):
            word_spans = data_batch["Token"]["span"][idx]
            predictions: Prediction = []
            for pred_idx, pred_args in srl_spans.items():
                begin, end = word_spans[pred_idx]
                # TODO cannot create annotation here.
                # Need to convert from Numpy numbers to int.
                pred_span = Span(begin.item(), end.item())
                arguments = []
                for arg in pred_args:
                    begin = word_spans[arg.start][0].item()
                    end = word_spans[arg.end][1].item()
                    arg_annotation = Span(begin, end)
                    arguments.append((arg_annotation, arg.label))
                predictions.append((pred_span, arguments))
            batch_predictions.append(predictions)
        return {"predictions": batch_predictions}

    def pack(
        self,
        pack: DataPack,
        predict_results: Dict[str, List[Prediction]],
        _: Optional[Annotation] = None,
    ):
        batch_predictions = predict_results["predictions"]
        for predictions in batch_predictions:
            for pred_span, arg_result in predictions:

                pred = PredicateMention(pack, pred_span.begin, pred_span.end)

                for arg_span, label in arg_result:
                    arg = PredicateArgument(pack, arg_span.begin, arg_span.end)
                    link = PredicateLink(pack, pred, arg)
                    link.arg_type = label

    @classmethod
    def default_configs(cls):
        """
        This defines the default configuration structure for the predictor.
        """
        return {
            "storage_path": None,
            "batcher": {
                "batch_size": 4,
                "context_type": "ft.onto.base_ontology.Sentence",
                "requests": {"ft.onto.base_ontology.Token": []},
            },
        }



# Copyright 2021 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A processor to delete overlapping annotations."""

__all__ = [
    "DeleteOverlapEntry",
]

from typing import List

from forte.common import Resources
from forte.common.configuration import Config
from forte.common.exception import ProcessorConfigError
from forte.data import Span
from forte.data.data_pack import DataPack
from forte.data.ontology import Annotation
from forte.processors.base import PackProcessor
from forte.utils import get_class


class DeleteOverlapEntry(PackProcessor):
    """
    A processor to delete overlapping annotations in a data pack. When
    overlapping, the first annotation (based on the iteration order) will be
    kept and the rest of them will be removed.
    """

    # pylint: disable=attribute-defined-outside-init,unused-argument
    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        if not self.configs.entry_type:
            raise ProcessorConfigError("Please specify an entity mention type!")

        self.entry_type = get_class(self.configs.entry_type)

        if not issubclass(self.entry_type, Annotation):
            raise AttributeError(
                f"The entry type to delete [{self.entry_type}] "
                f"is not a sub-class of "
                f"'forte.data.ontology.top.Annotation' class."
            )

    def _process(self, input_pack: DataPack):
        entry_spans: List[Span] = []
        entries: List[Annotation] = list(input_pack.get(self.entry_type))
        for entry in entries:
            current_span = entry.span
            if entry_spans and self._is_overlap(entry_spans[-1], current_span):
                input_pack.delete_entry(entry)
            else:
                entry_spans.append(current_span)

    def _is_overlap(self, interval1: Span, interval2: Span) -> bool:
        """
        Determine whether two intervals overlap with each other.
        """
        if interval2.begin < interval1.begin:
            interval1, interval2 = interval2, interval1
        return interval2.begin <= interval1.end

    @classmethod
    def default_configs(cls):
        """
        The ``entry_type`` config determines which type of annotation to be
        checked for duplication. This value should be the name of a class that
        is sub-class for :class:`~forte.data.ontology.top.Annotation`.
        Otherwise a `ValueError` will be raised.

        Returns:
            None.
        """

        return {
            "entry_type": None,
        }



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Utility functions related to data processing input/output.
"""
import os
from typing import Dict, List, Iterator, Any, Tuple

from forte.data.types import ReplaceOperationsType
from forte.data.span import Span

__all__ = [
    "batch_instances",
    "merge_batches",
    "slice_batch",
    "dataset_path_iterator",
]


def batch_instances(instances: List[Dict]):
    r"""Merge a list of ``instances``."""
    batch: Dict[str, Any] = {}
    for instance in instances:
        for entry, fields in instance.items():
            if isinstance(fields, dict):
                if entry not in batch.keys():
                    batch[entry] = {}
                for k, value in fields.items():
                    if k not in batch[entry].keys():
                        batch[entry][k] = []
                    batch[entry][k].append(value)
            else:  # context level feature
                if entry not in batch.keys():
                    batch[entry] = []
                batch[entry].append(fields)
    return batch


def merge_batches(batches: List[Dict]):
    r"""Merge a list of ``batches``."""
    merged_batch: Dict = {}
    for batch in batches:
        for entry, fields in batch.items():
            if isinstance(fields, dict):
                if entry not in merged_batch.keys():
                    merged_batch[entry] = {}
                for k, value in fields.items():
                    if k not in merged_batch[entry].keys():
                        merged_batch[entry][k] = []
                    merged_batch[entry][k].extend(value)
            else:  # context level feature
                if entry not in merged_batch.keys():
                    merged_batch[entry] = []
                merged_batch[entry].extend(fields)
    return merged_batch


def slice_batch(batch, start, length):
    r"""Return a sliced batch of size ``length`` from ``start`` in ``batch``."""
    sliced_batch: Dict = {}

    for batch_key, fields in batch.items():
        if isinstance(fields, dict):
            if batch_key not in sliced_batch.keys():
                sliced_batch[batch_key] = {}
            for k, value in fields.items():
                sliced_batch[batch_key][k] = value[start : start + length]
        else:  # context level feature
            sliced_batch[batch_key] = fields[start : start + length]

    return sliced_batch


def dataset_path_iterator_with_base(
    dir_path: str, file_extension: str
) -> Iterator[Tuple[str, str]]:
    r"""An iterator returning file_paths in a directory containing files of the
    given datasets, including the original directory as the first element.
    """
    for root, _, files in os.walk(dir_path):
        for data_file in files:
            if len(file_extension) > 0:
                if data_file.endswith(file_extension):
                    yield dir_path, os.path.join(root, data_file)
            else:
                yield dir_path, os.path.join(root, data_file)


def dataset_path_iterator(dir_path: str, file_extension: str) -> Iterator[str]:
    r"""An iterator returning the file paths in a directory containing files of
    the given datasets.
    """
    if not os.path.exists(dir_path):
        raise FileNotFoundError("Cannot find the directory [%s]." % dir_path)

    for root, _, files in os.walk(dir_path):
        for data_file in files:
            if len(file_extension) > 0:
                if data_file.endswith(file_extension):
                    yield os.path.join(root, data_file)
            else:
                yield os.path.join(root, data_file)


def modify_text_and_track_ops(
    original_text: str, replace_operations: ReplaceOperationsType
) -> Tuple[str, ReplaceOperationsType, List[Tuple[Span, Span]], int]:
    r"""Modifies the original text using ``replace_operations`` provided by the
    user to return modified text and other data required for tracking original
    text.

    Args:
        original_text: Text to be modified.
        replace_operations: A list of spans and the corresponding replacement
            string that the span in the original string is to be replaced with
            to obtain the original string.

    Returns:
        modified_text: Text after modification.
        replace_back_operations: A list of spans and the corresponding
            replacement string that the span in the modified string is to be
            replaced with to obtain the original string.
        processed_original_spans: List of processed span and its corresponding
            original span.
        orig_text_len: length of original text.
    """
    orig_text_len: int = len(original_text)
    mod_text: str = original_text
    increment: int = 0
    prev_span_end: int = 0
    replace_back_operations: List[Tuple[Span, str]] = []
    processed_original_spans: List[Tuple[Span, Span]] = []

    # Sorting the spans such that the order of replacement strings
    # is maintained -> utilizing the stable sort property of python sort
    replace_operations.sort(key=lambda item: item[0])

    for span, replacement in replace_operations:
        if span.begin < 0 or span.end < 0:
            raise ValueError("Negative indexing not supported")
        if span.begin > len(original_text) or span.end > len(original_text):
            raise ValueError(
                "One of the span indices are outside the string length"
            )
        if span.end < span.begin:
            print(span.begin, span.end)
            raise ValueError(
                "One of the end indices is lesser than start index"
            )
        if span.begin < prev_span_end:
            raise ValueError(
                "The replacement spans should be mutually exclusive"
            )
        span_begin = span.begin + increment
        span_end = span.end + increment
        original_span_text = mod_text[span_begin:span_end]
        mod_text = mod_text[:span_begin] + replacement + mod_text[span_end:]
        increment += len(replacement) - (span.end - span.begin)
        replacement_span = Span(span_begin, span_begin + len(replacement))
        replace_back_operations.append((replacement_span, original_span_text))
        processed_original_spans.append((replacement_span, span))
        prev_span_end = span.end

    return (
        mod_text,
        replace_back_operations,
        sorted(processed_original_spans),
        orig_text_len,
    )



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import KeysView
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import pickle

__all__ = ["Resources"]

SerializeDict = Dict[str, Callable[[Any, Union[str, Path]], None]]
DeserializeDict = Dict[str, Callable[[Union[str, Path]], None]]


class Resources:
    r"""The :class:`~forte.common.resources.Resources` object is a global
    registry used in the pipeline.
    Objects defined as :class:`~forte.common.resources.Resources` will be
    passed on to the processors in the
    pipeline for initialization.
    """

    def __init__(self, **kwargs):
        self._resources = {}
        self.update(**kwargs)

    def save(
        self,
        keys: Optional[Union[List[str], SerializeDict]] = None,
        output_dir: Optional[str] = None,
    ):
        r"""Save the resources specified by :attr:`keys` in binary format.

        Args:
            keys: list or dict:

                - If :attr:`keys` is a list, the objects corresponding to those keys
                  are saved
                - If :attr:`keys` is a dict mapping from a key to a serialize
                  function, then the serialize function will be used to save
                  the object corresponding to that key
                - If :attr:`keys` is None, all objects in this resource will be
                  saved.
            output_dir:
                A directory specifying the location to save the resources.
        """

        # TODO: use a default save directory like default_save_dir() if None
        if output_dir is None:
            output_dir = "./"

        if keys is None:
            keys = list(self._resources.keys())

        # pylint: disable=isinstance-second-argument-not-valid-type
        # TODO: disable until fix: https://github.com/PyCQA/pylint/issues/3507
        if isinstance(keys, List):
            for key in keys:
                with open(Path(output_dir, f"{key}.pkl"), "wb") as f:
                    pickle.dump(
                        self._resources.get(key), f, pickle.HIGHEST_PROTOCOL
                    )
        else:
            for key, serializer in keys.items():
                serializer(self._resources[key], Path(output_dir, f"{key}.pkl"))

    def keys(self) -> KeysView:
        r"""Return all keys of the resources."""
        return self._resources.keys()

    def contains(self, key: str) -> bool:
        """Return whether the specified key exists."""
        return key in self._resources.keys()

    def get(self, key: str):
        r"""Get the corresponding resource by specifying the key."""
        return self._resources.get(key)

    def update(self, **kwargs):
        r"""Update the resources."""
        self._resources.update(**kwargs)

    def remove(self, key: str):
        r"""Remove the corresponding resource by specifying the key."""
        del self._resources[key]

    def load(
        self,
        keys: Union[List[str], DeserializeDict],
        path: Optional[str] = None,
    ):
        r"""Load the resources specified by :attr:`keys`.

        Args:
            keys list or dict:

                - If :attr:`keys` is a list, the objects corresponding to those keys
                  are loaded
                - If :attr:`keys` is a dict mapping from a key to a deserialize
                  function, then the deserialize function will be used to load
                  the object corresponding to that key
            path: str
                A directory specifying the location to load the resources from.
        """

        # TODO: use a default save directory like default_save_dir() if None
        if path is None:
            path = "./"

        # pylint: disable=isinstance-second-argument-not-valid-type
        # TODO: disable until fix: https://github.com/PyCQA/pylint/issues/3507
        if isinstance(keys, List):
            for key in keys:
                with open(Path(path, f"{key}.pkl"), "rb") as f:
                    self._resources[key] = pickle.load(f)
        else:
            for key, deserializer in keys.items():
                self._resources[key] = deserializer(Path(path, f"{key}.pkl"))



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Defines the Evaluator interface and related functions.
"""
import logging
from abc import abstractmethod
from typing import Any, Dict, Set

from forte.data.base_pack import PackType
from forte.pipeline_component import PipelineComponent
from forte.utils.utils_processor import (
    record_types_and_attributes_check,
    collect_input_pack_record,
)

__all__ = [
    "Evaluator",
]


class Evaluator(PipelineComponent[PackType]):
    r"""The base class of the evaluator."""

    def __init__(self):
        super().__init__()
        self._pred_pack_expectation: Dict[str, Set[str]] = None
        self._ref_pack_expectation: Dict[str, Set[str]] = None

    @abstractmethod
    def consume_next(self, pred_pack: PackType, ref_pack: PackType):
        r"""The actual consume function that will be called by the pipeline.
        This function will deal with the basic pipeline status and call the
        `consume_next` function.

        Args:
            pred_pack: The prediction pack, which should contain the system
                predicted results.
            ref_pack: The reference pack, which should contain the reference
                to score on.
        """

        raise NotImplementedError

    @abstractmethod
    def get_result(self) -> Any:
        r"""The evaluator gather the results and the score should be obtained
        here.
        """
        raise NotImplementedError

    def expected_types_and_attributes(
        self,
        pred_pack_expectation: Dict[str, Set[str]],
        ref_pack_expectation: Dict[str, Set[str]],
    ):
        r"""If the evaluator has required input types and attributes for
        `pred_pack` or `ref_pack`, user could add the types and attributes
        required with this function.

        Args:
            pred_pack_expectation: The expected types and attributes of
                prediction pack.
            ref_pack_expectation: The expected types and attributes of
                reference pack.
        """
        self._pred_pack_expectation = pred_pack_expectation
        self._ref_pack_expectation = ref_pack_expectation

    def pred_pack_record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of prediction datapack of
        current processor to :attr:`forte.data.base_pack.BaseMeta.record`.

        Args:
            record_meta: The field in the datapack for type record that need to
                fill in for consistency checking.
        """
        pass

    def ref_pack_record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of reference datapack of
        current processor to :attr:`forte.data.base_pack.BaseMeta.record`.

        Args:
            record_meta: The field in the datapack for record that need to
                fill in for consistency checking.
        """
        pass

    def check_record(self, pred_pack: PackType, ref_pack: PackType):
        # pylint: disable=protected-access
        r"""Method to check type consistency if
        :attr:`~forte.pipeline.Pipeline.enforce_consistency` is enabled
        for the pipeline. If any expected type or its attribute
        does not exist in the `pred_pack` or `ref_pack` record of the previous
        pipeline component, an error of
        :class:`~forte.common.exception.ExpectedRecordNotFound` will be raised.

        Args:
            pred_pack: The prediction pack, which should contain the system
                predicted results.
            ref_pack: The reference pack, which should contain the reference
                to score on.
        """
        if self._check_type_consistency:
            pred_pack_record = collect_input_pack_record(
                self.resources, pred_pack
            )
            ref_pack_record = collect_input_pack_record(
                self.resources, ref_pack
            )
            record_types_and_attributes_check(
                self._pred_pack_expectation, pred_pack_record
            )
            record_types_and_attributes_check(
                self._ref_pack_expectation, ref_pack_record
            )

    def writes_record(self, pred_pack: PackType, ref_pack: PackType):
        r"""Method to write records of the output type of the current
        processor to the datapack. The key of the record should be the entry
        type and values should be attributes of the entry type. All the
        information would be used for consistency checking purpose if the
        pipeline is initialized with
        `enforce_consistency=True`.

        Args:
            pred_pack: The prediction pack, which should contain the system
                predicted results.
            ref_pack: The reference pack, which should contain the reference
                to score on.

        """
        # pylint: disable=protected-access
        if self._check_type_consistency:
            try:
                self.pred_pack_record(pred_pack._meta.record)
            except AttributeError:
                # For backward compatibility, no record to write.
                logging.info(
                    "Packs of the old format do not have the record field."
                )

            try:
                self.ref_pack_record(ref_pack._meta.record)
            except AttributeError:
                # For backward compatibility, no record to write.
                logging.info(
                    "Packs of the old format do not have the record field."
                )



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads RACE multi choice QA data into Datapacks.
"""
import json
import os
from typing import Any, Iterator

from forte.data.data_pack import DataPack
from forte.data.data_utils_io import dataset_path_iterator
from forte.data.base_reader import PackReader
from ft.onto.base_ontology import MCOption, MCQuestion, Document
from ftx.onto.race_qa import Passage


__all__ = [
    "RACEMultiChoiceQAReader",
]


class RACEMultiChoiceQAReader(PackReader):
    r""":class:`RACEMultiChoiceQAReader` is designed to read in RACE multi
    choice qa dataset.
    """

    def _collect(self, json_directory) -> Iterator[Any]:  # type: ignore
        r"""Should be called with param ``json_directory`` which is a path to a
        folder containing json files.

        Args:
            json_directory: directory containing the json files.

        Returns: Iterator over paths to .json files
        """
        return dataset_path_iterator(json_directory, "")

    def _cache_key_function(self, json_file: str) -> str:
        return os.path.basename(json_file)

    def _convert_to_int(self, ch: Any) -> int:
        if isinstance(ch, int):
            return ch
        if isinstance(ch, str):
            return ord(ch.lower()) - 97
        raise ValueError(
            "Wrong datatype for Answers: expected int or str, "
            f"got {type(ch).__name__}"
        )

    def _parse_pack(self, file_path: str) -> Iterator[DataPack]:
        with open(file_path, "r", encoding="utf8", errors="ignore") as file:
            dataset = json.load(file)

            pack = DataPack()
            text: str = dataset["article"]
            article_end = len(text)
            offset = article_end + 1

            for qid, ques_text in enumerate(dataset["questions"]):
                text += "\n" + ques_text
                ques_end = offset + len(ques_text)
                question = MCQuestion(pack, offset, ques_end)
                offset = ques_end + 1

                options_text = dataset["options"][qid]
                for option_text in options_text:
                    text += "\n" + option_text
                    option_end = offset + len(option_text)
                    option = MCOption(pack, offset, option_end)
                    offset = option_end + 1
                    question.options.append(option)

                answers = dataset["answers"][qid]
                if not isinstance(answers, list):
                    answers = [answers]
                answers = [self._convert_to_int(ans) for ans in answers]
                question.answers = answers

            pack.set_text(text, replace_func=self.text_replace_operation)

            Document(pack, 0, article_end)

            passage_id: str = dataset["id"]
            passage = Passage(pack, 0, len(pack.text))
            passage.passage_id = passage_id

            pack.pack_name = passage_id
            yield pack



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads SemEval Task8 data with annotations into Datapacks.
"""

import os
from typing import Any, Iterator, Dict
from forte.data.data_pack import DataPack
from forte.data.base_reader import PackReader
from forte.data.data_utils_io import dataset_path_iterator
from ft.onto.base_ontology import Sentence, RelationLink, EntityMention

__all__ = ["SemEvalTask8Reader"]


class SemEvalTask8Reader(PackReader):
    r""":class:`SemEvalTask8Reader` is designed to read in
        SemEval Task-8 dataset. The data can be obtained
        here: http://www.kozareva.com/downloads.html


        `Hendrickx, Iris, et al. SemEval-2010 task 8: Multi-way
        classification of semantic relations between pairs of
        nominals.` https://www.aclweb.org/anthology/S10-1006.pdf

        An example of the dataset is
        '''
        8	"<e1>People</e1> have been moving back \
        into <e2>downtown</e2>."
        Entity-Destination(e1,e2)
        Comment:
        '''.

        This example will be converted to one `Sentence`,
        "People have been moving back into downtown."
        and one RelationLink,
        link = RelationLink(parent=People, child=downtown)
        link.rel_type = Entity-Destination
        into the DataPack.
    """

    def _cache_key_function(self, file_path: str) -> str:
        return os.path.basename(file_path)

    def _collect(self, *args, **kwargs) -> Iterator[Any]:
        # pylint: disable = unused-argument
        r"""args[0] should be the folder where
        the SemEval Task8 dataset is stored.
        Files ended with sem_eval_task8_file_extension (.txt)
        are exptected here.

        Args:
            args: args[0] is the directory to the dataset.

        Returns: Iterator over the file name (str).
        """
        sem_file_dir: str = args[0]
        return dataset_path_iterator(
            sem_file_dir, self.configs.sem_eval_task8_file_extension
        )

    def _parse_pack(self, file_path: str) -> Iterator[DataPack]:
        pack: DataPack = DataPack()

        with open(file_path, "r", encoding="utf8") as fp:
            txt = ""
            offset = 0

            while True:
                sent_line: str = fp.readline()
                if not sent_line:
                    break

                if len(sent_line.split()) == 0:
                    continue

                relation_line: str = fp.readline()
                # Command line is not used.
                _ = fp.readline()

                sent_line = sent_line[
                    sent_line.find('"') + 1 : sent_line.rfind('"')
                ]
                index1 = sent_line.find("<e1>")
                index2 = sent_line.find("<e2>")
                # 5 is the length of "</e1>", include both <e1> and
                # </e1> when extracting the string.
                e1 = sent_line[index1 : sent_line.find("</e1>") + 5]
                e2 = sent_line[index2 : sent_line.find("</e2>") + 5]
                # Remove <e1> and </e1> in the sentence.
                sent_line = sent_line.replace(e1, e1[4:-5])
                sent_line = sent_line.replace(e2, e2[4:-5])
                # Remove <e1> and </e1> in e1.
                e1 = e1[4:-5]
                e2 = e2[4:-5]
                # Re-calculate the index after removing <e1>, </e1> in
                # in the sentence.
                if index1 < index2:
                    diff1 = 0
                    diff2 = 9
                else:
                    diff1 = 9
                    diff2 = 0
                index1 += offset - diff1
                index2 += offset - diff2

                Sentence(pack, offset, offset + len(sent_line))
                entry1 = EntityMention(pack, index1, index1 + len(e1))
                entry2 = EntityMention(pack, index2, index2 + len(e2))
                offset += len(sent_line) + 1
                txt += sent_line + " "

                pair = relation_line[
                    relation_line.find("(") + 1 : relation_line.find(")")
                ]

                if "," in pair:
                    parent, _ = pair.split(",")
                    if parent == "e1":
                        relation = RelationLink(pack, entry1, entry2)
                    else:
                        relation = RelationLink(pack, entry2, entry1)
                    relation.rel_type = relation_line[: relation_line.find("(")]
                else:
                    # For "Other" relation, just set parent as e1
                    # set child as e2.
                    relation = RelationLink(pack, entry1, entry2)
                    relation.rel_type = relation_line.strip()

        pack.set_text(txt, replace_func=self.text_replace_operation)
        pack.pack_name = os.path.basename(file_path)

        yield pack

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        configs: Dict = super().default_configs()

        configs.update({"sem_eval_task8_file_extension": "txt"})
        return configs



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A wrapper adding data augmentation to a Bert model with arbitrary tasks.
"""

import random
import math
from typing import Tuple, Dict, Generator
from forte.utils import create_import_error_msg
from forte.models.da_rl.magic_model import MetaModule

try:
    import torch
    from torch import nn
    from torch.nn import functional as F
    from torch.optim import Optimizer
except ImportError as e:
    raise ImportError(
        create_import_error_msg("torch", "models", "Augmentation Wrapper")
    ) from e


try:
    import texar.torch as tx
except ImportError as e:
    raise ImportError(
        create_import_error_msg(
            "texar-pytorch", "models", "Augmentation Wrapper"
        )
    ) from e

__all__ = ["MetaAugmentationWrapper"]


class MetaAugmentationWrapper:
    # pylint: disable=line-too-long
    r"""
    A wrapper adding data augmentation to a Bert model with arbitrary tasks.
    This is used to perform reinforcement learning for joint data augmentation
    learning and model training.

    See: https://arxiv.org/pdf/1910.12795.pdf

    This code is adapted from:
    https://github.com/tanyuqian/learning-data-manipulation/blob/master/augmentation/generator.py

    Let :math:`\theta` be the parameters of the downstream (classifier) model.
    Let :math:`\phi` be the parameters of the augmentation model.
    Equations to update :math:`\phi`:

    .. math::

        \theta'(\phi) = \theta - \nabla_{\theta} L_{train}(\theta, \phi)

        \phi = \phi - \nabla_{\phi} L_{val}(\theta'(\phi))

    Args:
        augmentation_model:
            A Bert-based model for data augmentation.
            E.g. BertForMaskedLM.
            Model requirement: masked language modeling, the output logits
            of this model is of shape `[batch_size, seq_length, token_size]`.
        augmentation_optimizer:
            An optimizer that is associated with `augmentation_model`.
            E.g. Adam optimizer.
        input_mask_ids:
            Bert token id of `'[MASK]'`. This is used to randomly mask out
            tokens from the input sentence during training.
        device:
            The CUDA device to run the model on.
        num_aug:
            The number of samples from the augmentation model
            for every augmented training instance.

    Example usage:

        .. code-block:: python

            aug_wrapper = MetaAugmentationWrapper(
                aug_model, aug_optim, mask_id, device, num_aug)
            for batch in training_data:
                # Train augmentation model params.
                aug_wrapper.reset_model()
                for instance in batch:
                    # Augmented example with params phi exposed
                    aug_instance_features = \
                        aug_wrapper.augment_instance(instance_features)
                    # Model is the downstream Bert model.
                    model.zero_grad()
                    loss = model(aug_instance_features)
                    meta_model = MetaModule(model)
                    meta_model = aug_wrapper.update_meta_model(
                        meta_model, loss, model, optim)

                    # Compute gradient of the augmentation model on validation
                    data for val_batch in validation_data:
                        val_loss = meta_model(val_batch_features)
                        val_loss = val_loss / num_training_instance / num_aug \
                            / num_val_batch
                        val_loss.backward()
                # update augmentation model params.
                aug_wrapper.update_phi()

                # train classifier with augmented batch
                aug_batch_features = aug_wrapper.augment_batch(batch_features)
                optim.zero_grad()
                loss = model(aug_batch_features)
                loss.backward()
                optim.step()

    """

    def __init__(
        self,
        augmentation_model: nn.Module,
        augmentation_optimizer: Optimizer,
        input_mask_ids: int,
        device: torch.device,
        num_aug: int,
    ):
        self._aug_model = augmentation_model
        self._aug_optimizer = augmentation_optimizer
        self._input_mask_ids = input_mask_ids
        self._device = device
        self._num_aug = num_aug

    def reset_model(self):
        self._aug_model.train()
        self._aug_model.zero_grad()

    def _augment_instance(
        self, features: Tuple[torch.Tensor, ...], num_aug: int
    ) -> torch.Tensor:
        r"""Augment a training instance. Randomly mask out some tokens in the
        input sentence and use the logits of the augmentation model as the
        augmented bert token soft embedding.

        Args:
            features: A tuple of Bert features of one training instance.
                `(input_ids, input_mask, segment_ids, label_ids)`.

                `input_ids` is a tensor of Bert token ids.
                It has shape `[seq_len]`.

                `input_mask` is a tensor of shape `[seq_len]` with 1 indicating
                without mask and 0 with mask.

                `segment_ids` is a tensor of shape `[seq_len]`.
                `label_ids` is a tensor of shape `[seq_len]`.
            num_aug: The number of samples from the augmentation model.

        Returns:
            aug_probs: A tensor of shape `[num_aug, seq_len, token_size]`.
                It is the augmented bert token soft embedding.
        """

        feature: Generator[torch.Tensor, torch.Tensor, torch.Tensor] = (
            t.view(1, -1).to(self._device) for t in features
        )
        init_ids, input_mask, segment_ids, _ = feature

        length = int(torch.sum(input_mask).item())

        if length >= 4:
            mask_idx = sorted(
                random.sample(list(range(1, length - 1)), max(length // 7, 2))
            )
        else:
            mask_idx = [1]

        init_ids[0][mask_idx] = self._input_mask_ids

        logits = self._aug_model(
            init_ids, token_type_ids=segment_ids, attention_mask=input_mask
        )[0]

        # Get samples
        aug_probs_all = []
        for _ in range(num_aug):
            # Need a gumbel trick here in order to keep phi as variables.
            # Enable efficient gradient propagation through theta' to phi.
            probs = F.gumbel_softmax(logits.squeeze(0), hard=False)
            aug_probs = torch.zeros_like(probs).scatter_(
                1, init_ids[0].unsqueeze(1), 1.0
            )

            for t in mask_idx:
                aug_probs = tx.utils.pad_and_concat(
                    [aug_probs[:t], probs[t : t + 1], aug_probs[t + 1 :]],
                    axis=0,
                )

            aug_probs_all.append(aug_probs)

        aug_probs = tx.utils.pad_and_concat(
            [ap.unsqueeze(0) for ap in aug_probs_all], axis=0
        )

        return aug_probs

    def augment_instance(
        self, features: Tuple[torch.Tensor, ...]
    ) -> Tuple[torch.Tensor, ...]:
        r"""Augment a training instance.

        Args:
            features: A tuple of Bert features of one training instance.
                `(input_ids, input_mask, segment_ids, label_ids)`.

                `input_ids` is a tensor of Bert token ids.
                It has shape `[seq_len]`.

                `input_mask` is a tensor of shape `[seq_len]` with 1 indicating
                without mask and 0 with mask.

                `segment_ids` is a tensor of shape `[seq_len]`.
                `label_ids` is a tensor of shape `[seq_len]`.

        Returns:
            A tuple of Bert features of augmented training instances.
            `(input_probs_aug, input_mask_aug, segment_ids_aug, label_ids_aug)`.

            `input_probs_aug` is a tensor of soft Bert embeddings,
            distributions over vocabulary.
            It has shape `[num_aug, seq_len, token_size]`.
            It keeps :math:`\phi` as variable so that after passing it as an
            input to the classifier, the gradients of :math:`\theta` will
            also apply to :math:`\phi`.

            `input_mask_aug` is a tensor of shape `[num_aug, seq_len]`, it
            concatenates `num_aug` the input `input_mask` so that it
            corresponds to the mask of each token in `input_probs_aug`.

            `segment_ids_aug` is a tensor of shape `[num_aug, seq_len]`, it
            concatenates `num_aug` the input `segment_ids` so that it
            corresponds to the token type of each token in `input_probs_aug`.

            `label_ids_aug` is a tensor of shape `[num_aug, seq_len]`, it
            concatenates `num_aug` the input `label_ids` so that it corresponds
            to the label of each token in `input_probs_aug`.
        """

        aug_probs = self._augment_instance(features, self._num_aug)

        _, input_mask, segment_ids, label_ids = (
            t.to(self._device).unsqueeze(0) for t in features
        )
        input_mask_aug = tx.utils.pad_and_concat(
            [input_mask] * self._num_aug, axis=0
        )
        segment_ids_aug = tx.utils.pad_and_concat(
            [segment_ids] * self._num_aug, axis=0
        )
        label_ids_aug = tx.utils.pad_and_concat(
            [label_ids] * self._num_aug, axis=0
        )

        return aug_probs, input_mask_aug, segment_ids_aug, label_ids_aug

    def augment_batch(
        self, batch_features: Tuple[torch.Tensor, ...]
    ) -> Tuple[torch.Tensor, ...]:
        r"""Augment a batch of training instances. Append augmented instances
        to the input instances.

        Args:
            batch_features: A tuple of Bert features of a batch training
                instances. (input_ids, input_mask, segment_ids, label_ids).

                `input_ids` is a tensor of Bert token ids.
                It has shape `[batch_size, seq_len]`.

                `input_mask`, `segment_ids`, `label_ids` are all tensors of
                shape `[batch_size, seq_len]`.

        Returns:
            A tuple of Bert features of augmented training instances.
            `(input_probs_aug, input_mask_aug, segment_ids_aug, label_ids_aug)`.

            `input_probs_aug` is a tensor of soft Bert embeddings,
            It has shape `[batch_size * 2, seq_len, token_size]`.

            `input_mask_aug` is a tensor of shape `[batch_size * 2, seq_len]`,
            it concatenates two input `input_mask`, the first one corresponds to the
            mask of the tokens in the original bert instance, the second one
            corresponds to the mask of the augmented bert instance.

            `segment_ids_aug` is a tensor of shape `[batch_size * 2, seq_len]`,
            it concatenates two input `segment_ids`, the first one corresponds
            to the segment id of the tokens in the original bert instance, the
            second one corresponds to the segment id of the
            augmented bert instance.

            `label_ids_aug` is a tensor of shape `[batch_size * 2, seq_len]`,
            it concatenates two input `label_ids`, the first one corresponds
            to the labels of the original bert instance, the second one
            corresponds to the labels of the augmented bert instance.
        """
        input_ids, input_mask, segment_ids, labels = batch_features
        self._aug_model.eval()

        aug_instances = []
        features = []
        num_instance = len(input_ids)
        for i in range(num_instance):
            feature = (input_ids[i], input_mask[i], segment_ids[i], labels[i])
            features.append(feature)
            with torch.no_grad():
                aug_probs = self._augment_instance(feature, num_aug=1)
                aug_instances.append(aug_probs)

        input_ids_or_probs, input_masks, segment_ids, label_ids = [
            tx.utils.pad_and_concat(
                [t[i].unsqueeze(0) for t in features], axis=0
            ).to(self._device)
            for i in range(4)
        ]

        num_aug = len(aug_instances[0])

        input_ids_or_probs_aug = []
        for i in range(num_aug):
            for aug_probs in aug_instances:
                input_ids_or_probs_aug.append(aug_probs[i : i + 1])
        input_ids_or_probs_aug = tx.utils.pad_and_concat(
            input_ids_or_probs_aug, axis=0
        ).to(self._device)

        inputs_onehot = torch.zeros_like(
            input_ids_or_probs_aug[: len(input_ids_or_probs)]
        ).scatter_(2, input_ids_or_probs.unsqueeze(2), 1.0)
        input_probs_aug = tx.utils.pad_and_concat(
            [inputs_onehot, input_ids_or_probs_aug], axis=0
        ).to(self._device)

        input_mask_aug = tx.utils.pad_and_concat(
            [input_masks] * (num_aug + 1), axis=0
        ).to(self._device)
        segment_ids_aug = tx.utils.pad_and_concat(
            [segment_ids] * (num_aug + 1), axis=0
        ).to(self._device)
        label_ids_aug = tx.utils.pad_and_concat(
            [label_ids] * (num_aug + 1), axis=0
        ).to(self._device)

        return input_probs_aug, input_mask_aug, segment_ids_aug, label_ids_aug

    def eval_batch(
        self, batch_features: Tuple[torch.Tensor, ...]
    ) -> torch.FloatTensor:
        r"""Evaluate a batch of training instances.

        Args:
            batch_features: A tuple of Bert features of a batch training
                instances. (input_ids, input_mask, segment_ids, label_ids).

                `input_ids` is a tensor of Bert token ids.
                It has shape `[batch_size, seq_len]`.

                `input_mask`, `segment_ids`, `label_ids` are all tensors of
                shape `[batch_size, seq_len]`.

        Returns:
            The masked language modeling loss of one evaluation batch.
            It is a `torch.FloatTensor` of shape `[1,]`.
        """
        self._aug_model.eval()
        batch = tuple(t.to(self._device) for t in batch_features)
        input_ids, input_mask, segment_ids, labels = batch
        loss = self._aug_model(
            input_ids,
            token_type_ids=segment_ids,
            attention_mask=input_mask,
            labels=labels,
        )[0]
        return loss

    def update_meta_model(
        self,
        meta_model: MetaModule,
        loss: torch.Tensor,
        model: nn.Module,
        optimizer: Optimizer,
    ) -> MetaModule:
        r"""Update the parameters within the `MetaModel`
        according to the downstream model loss.

        `MetaModel` is used to calculate
        :math:`\nabla_{\phi} L_{val}(\theta'(\phi))`,
        where it needs gradients applied to :math:`\phi`.

        Perform parameter updates in this function, and later applies gradient
        change to :math:`\theta` and :math:`\phi` using validation data.

        Args:
            meta_model: A meta model whose parameters will be updated in-place
                by the deltas calculated from the input `loss`.
            loss: The loss of the downstream model that have taken
                the augmented training instances as input.
            model: The downstream Bert model.
            optimizer: The optimizer that is associated with the `model`.

        Returns:
            The same input `meta_model` with the updated parameters.
        """

        # grads_theta(phi) = \nabla_{theta} L_{train}(theta, phi)
        grads_theta = self._calculate_grads(loss, model, optimizer)

        # theta'(phi) = theta - grads_theta(phi)
        meta_model.update_params(grads_theta)

        return meta_model

    @staticmethod
    def _calculate_grads(
        loss: torch.Tensor, model: nn.Module, optimizer: Optimizer
    ) -> Dict[str, torch.Tensor]:
        grads = torch.autograd.grad(
            loss,
            [param for name, param in model.named_parameters()],
            create_graph=True,
        )
        grads = {
            param: grads[i]
            for i, (name, param) in enumerate(model.named_parameters())
        }

        if isinstance(optimizer, tx.core.BertAdam):
            deltas = _texar_bert_adam_delta(grads, model, optimizer)
        else:
            deltas = _torch_adam_delta(grads, model, optimizer)

        return deltas

    def update_phi(self):
        # L_{val}(theta'(phi))
        # apply gradients to phi

        # phi = phi - \nabla_{phi} L_{val}(theta'(phi))
        self._aug_optimizer.step()


def _texar_bert_adam_delta(
    grads: Dict[nn.parameter.Parameter, torch.Tensor],
    model: nn.Module,
    optimizer: Optimizer,
) -> Dict[str, torch.Tensor]:
    # pylint: disable=line-too-long
    r"""Compute parameter delta function for texar-pytorch
    core.BertAdam optimizer.

    This function is adapted from:
    https://github.com/asyml/texar-pytorch/blob/master/texar/torch/core/optimization.py#L398
    """
    assert isinstance(optimizer, tx.core.BertAdam)

    deltas = {}
    for group in optimizer.param_groups:
        for param in group["params"]:
            grad = grads[param]
            state = optimizer.state[param]

            if len(state) == 0:
                # Exponential moving average of gradient values
                state["next_m"] = torch.zeros_like(param.data)
                # Exponential moving average of squared gradient values
                state["next_v"] = torch.zeros_like(param.data)

            exp_avg, exp_avg_sq = state["next_m"], state["next_v"]

            beta1, beta2 = group["betas"]

            if group["weight_decay"] != 0:
                grad = grad + group["weight_decay"] * param.data

            exp_avg = exp_avg * beta1 + (1.0 - beta1) * grad
            exp_avg_sq = exp_avg_sq * beta2 + (1.0 - beta2) * grad * grad
            denom = exp_avg_sq.sqrt() + group["eps"]

            step_size = group["lr"]

            deltas[param] = -step_size * exp_avg / denom

    param_to_name = {param: name for name, param in model.named_parameters()}

    return {param_to_name[param]: delta for param, delta in deltas.items()}


def _torch_adam_delta(
    grads: Dict[nn.parameter.Parameter, torch.Tensor],
    model: nn.Module,
    optimizer: Optimizer,
) -> Dict[str, torch.Tensor]:
    r"""Compute parameter delta function for Torch Adam optimizer."""
    assert issubclass(type(optimizer), Optimizer)

    deltas = {}
    for group in optimizer.param_groups:
        for param in group["params"]:
            grad = grads[param]
            state = optimizer.state[param]

            if len(state) == 0:
                state["exp_avg"] = torch.zeros_like(param.data)
                state["exp_avg_sq"] = torch.zeros_like(param.data)
                state["step"] = 0

            exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
            beta1, beta2 = group["betas"]

            step = state["step"] + 1

            if group["weight_decay"] != 0:
                grad = grad + group["weight_decay"] * param.data

            exp_avg = exp_avg * beta1 + (1.0 - beta1) * grad
            exp_avg_sq = exp_avg_sq * beta2 + (1.0 - beta2) * grad * grad
            denom = exp_avg_sq.sqrt() + group["eps"]

            bias_correction1 = 1.0 - beta1**step
            bias_correction2 = 1.0 - beta2**step
            step_size = (
                group["lr"] * math.sqrt(bias_correction2) / bias_correction1
            )

            deltas[param] = -step_size * exp_avg / denom

    param_to_name = {param: name for name, param in model.named_parameters()}

    return {param_to_name[param]: delta for param, delta in deltas.items()}



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Base class for Pipeline module.
"""

import os
import itertools
import json
import logging
import sys
import uuid
from pathlib import Path
from time import time
from typing import (
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Union,
    Tuple,
    Deque,
    Set,
)

import yaml

from forte.common import ProcessorConfigError
from forte.common.configuration import Config
from forte.common.exception import (
    ProcessExecutionException,
    ProcessFlowException,
    ValidationError,
)
from forte.common.resources import Resources
from forte.data.base_pack import PackType
from forte.data.base_reader import BaseReader
from forte.data.caster import Caster
from forte.data.ontology.code_generation_objects import EntryTree
from forte.data.ontology.ontology_code_generator import OntologyCodeGenerator
from forte.data.selector import Selector, DummySelector
from forte.evaluation.base.base_evaluator import Evaluator
from forte.pipeline_component import PipelineComponent
from forte.process_job import ProcessJob
from forte.process_manager import ProcessManager, ProcessJobStatus
from forte.processors.base import BaseProcessor
from forte.processors.base.batch_processor import BaseBatchProcessor
from forte.utils import (
    create_class_with_kwargs,
    get_full_module_name,
    create_import_error_msg,
)
from forte.utils.utils_processor import record_types_and_attributes_check
from forte.version import FORTE_IR_VERSION

if sys.version_info < (3, 7):
    import importlib_resources as resources
else:
    from importlib import resources

logger = logging.getLogger(__name__)

__all__ = ["Pipeline", "serve"]


class ProcessBuffer:
    def __init__(self, pipeline: "Pipeline", data_iter: Iterator[PackType]):
        self.__data_iter: Iterator[PackType] = data_iter
        self.__data_exhausted = False
        self.__pipeline = pipeline
        self.__process_manager: ProcessManager = pipeline._proc_mgr

    def __iter__(self):
        return self

    def __next__(self) -> ProcessJob:
        if self.__process_manager.current_queue_index == -1:
            if self.__data_exhausted:
                # Both the buffer is empty and the data input is exhausted.
                raise StopIteration
            try:
                job_pack = next(self.__data_iter)
                job = ProcessJob(job_pack, False)

                if len(self.__pipeline.evaluator_indices) > 0:
                    gold_copy = job_pack.view()
                    self.__pipeline.add_gold_packs({job.id: gold_copy})

                self.__process_manager.add_to_queue(queue_index=0, job=job)
                self.__process_manager.current_queue_index = 0
                self.__process_manager.current_processor_index = 0
                return job
            except StopIteration:
                self.__data_exhausted = True
                job = ProcessJob(None, True)
                self.__process_manager.add_to_queue(queue_index=0, job=job)
                self.__process_manager.current_queue_index = 0
                self.__process_manager.current_processor_index = 0
                return job
        else:
            q_index = self.__process_manager.current_queue_index
            u_index = self.__process_manager.unprocessed_queue_indices[q_index]
            return self.__process_manager.current_queue[u_index]


class Pipeline(Generic[PackType]):
    # pylint: disable=too-many-public-methods
    r"""This controls the main inference flow of the system. A pipeline is
    consisted of a set of Components (readers and processors). The data flows
    in the pipeline as data packs, and each component will use or add
    information to the data packs.
    """

    def __init__(
        self,
        resource: Optional[Resources] = None,
        ontology_file: Optional[str] = None,
        enforce_consistency: bool = False,
        do_init_type_check: bool = False,
    ):
        r"""

        Args:
            resource: The :class:`Resources` object, which is a global registry used
                in the pipeline. Objects defined as :class:`Resources` will be
                passed on to the processors in the
                pipeline for initialization.
            ontology_file: The path to the input ontology specification file,
                which should be a json file, and it should have all the entries
                inside with no import as key.
            enforce_consistency: This boolean determines whether the
                pipeline will check the content expectations specified in each
                pipeline component. Each component will check whether the input
                pack contains the expected data
                via checking the meta-data, and throws a
                :class:`~forte.common.exception.EntryNotFoundError` if it
                fails. When this function is called with enforce is ``True``,
                all the pipeline components would check if the input datapack
                record matches
                with the expected types and attributes if function
                ``expected_types_and_attributes`` is implemented
                for the processor. For example, processor A requires entry type
                of :class:`~ft.onto.base_ontology.Sentence`, and processor B would
                produce this type in the output datapack, so ``record`` function
                of processor B writes the record of this type in the datapack
                and processor A implements ``expected_types_and_attributes`` to
                add this type. Then when the pipeline runs with
                `enforce_consistency=True`, processor A would check if this
                type exists in the record of the output of the
                previous pipeline component.
            do_init_type_check: Determine whether to check records types and
                attributes during pipeline initialization. Default to `False`.
                If this boolean is set to `True`, each component in the
                pipeline will be validated by comparing its
                ``expected_types_and_attributes`` with the accumulated
                ``records`` from all the downstream components.
        """
        self._reader: BaseReader
        self._reader_config: Optional[Config] = None

        # These variables defines the units in the pipeline, they should be
        # of the same length
        self._components: List[PipelineComponent] = []
        self._selectors: List[Selector] = []
        self._configs: List[Optional[Config]] = []
        self._selectors_configs: List[Optional[Config]] = []
        # corresponding to the new added parameter "ref_name", indicating a list of
        # reference names that are used to identify different components
        self._ref_names: Dict[str, Any] = {}
        # Maintain a set of the pipeline components to fast check whether
        # the component is already there.
        self.__component_set: Set[PipelineComponent] = set()

        # Will initialize at `initialize` because the processors length is
        # unknown.
        self._proc_mgr: ProcessManager = None  # type: ignore

        self.evaluator_indices: List[int] = []

        # needed for evaluator
        self._predict_to_gold: Dict[int, PackType] = {}

        if resource is None:
            self.resource = Resources()
        else:
            self.resource = resource

        if ontology_file is None:
            with resources.path(
                "forte.ontology_specs", "base_ontology.json"
            ) as data_path:
                ontology_file = str(data_path)

        if ontology_file is not None:
            with open(ontology_file, "r", encoding="ascii") as f:
                spec_dict = json.load(f)
                self.resource.update(onto_specs_path=ontology_file)
                self.resource.update(onto_specs_dict=spec_dict)

        # The flag indicating whether this pipeline is initialized.
        self._initialized: bool = False
        # The flag indicating whether we want to enforce type consistency
        #  between the processors.
        self._check_type_consistency: bool = False

        # Create one copy of the dummy selector to reduce class creation.
        self.__default_selector: Selector = DummySelector()
        self.__default_selector_config: Config = Config({}, {})

        # needed for time profiling of pipeline
        self._enable_profiling: bool = False
        self._profiler: List[float] = []

        self._check_type_consistency = enforce_consistency

        # Indicate whether do type checking during pipeline initialization
        self._do_init_type_check: bool = do_init_type_check

    def enforce_consistency(self, enforce: bool = True):
        r"""This function determines whether the pipeline will check
        the content expectations specified in each pipeline component. This
        function works with :meth:`~forte.pipeline.Pipeline.initialize` called
        after itself. Each component will check whether the input pack contains
        the expected data via checking the meta-data, and throws a
        :class:`~forte.common.exception.EntryNotFoundError` if the check
        fails. The example of implementation is mentioned in the docstrings of
        :meth:`~forte.pipeline.Pipeline.__init__`.

        Args:
            enforce: A boolean of whether to enable consistency checking
                for the pipeline or not.
        """
        self._check_type_consistency = enforce

    def init_from_config_path(self, config_path):
        r"""Read the configurations from the given path ``config_path``
        and build the pipeline with the config.

        Args:
            config_path: A string of the configuration path, which is
                is a YAML file that specify the structure and parameters of the
                pipeline.
        """
        with open(config_path, encoding="utf-8") as f:
            configs = yaml.safe_load(f)
            self.init_from_config(configs)

    def init_from_config(self, configs: Dict[str, Any]):
        r"""Initialized the pipeline (ontology and processors) from the
        given configurations.

        Args:
            configs: The configs used to initialize the pipeline. It should be
                a dictionary that contains `forte_ir_version`, ``components``
                and `states`. `forte_ir_version` is a string used to validate
                input format. ``components`` is a list of dictionary that
                contains `type` (the class of pipeline components),
                `configs` (the corresponding component's configs) and
                `selector`. `states` will be used to update the pipeline states
                based on the fields specified in `states.attribute` and
                `states.resource`.
        """
        # Validate IR version
        if configs.get("forte_ir_version") != FORTE_IR_VERSION:
            raise ProcessorConfigError(
                f"forte_ir_version={configs.get('forte_ir_version')} not "
                "supported. Please make sure the format of input IR complies "
                f"with forte_ir_version={FORTE_IR_VERSION}."
            )

        # Add components from IR
        is_first: bool = True
        for component_config in configs.get("components", []):
            component = create_class_with_kwargs(
                class_name=component_config["type"],
                class_args=component_config.get("kwargs", {}),
            )

            if is_first:
                if not isinstance(component, BaseReader):
                    raise ProcessorConfigError(
                        "The first component of a pipeline must be a reader."
                    )
                self.set_reader(component, component_config.get("configs", {}))
                is_first = False
            else:
                # Can be processor, caster, or evaluator
                selector_config = component_config.get("selector")
                self.add(
                    component=component,
                    config=component_config.get("configs", {}),
                    selector=selector_config
                    and create_class_with_kwargs(
                        class_name=selector_config["type"],
                        class_args=selector_config.get("kwargs", {}),
                    ),
                    selector_config=None
                    if selector_config is None
                    else selector_config.get("configs"),
                )

        # Set pipeline states and resources
        states_config: Dict[str, Dict] = configs.get("states", {})
        for attr, val in states_config.get("attribute", {}).items():
            setattr(self, attr, val)
        resource_config: Dict[str, Dict] = states_config.get("resource", {})
        if "onto_specs_dict" in resource_config:
            self.resource.update(
                onto_specs_dict=resource_config["onto_specs_dict"]
            )
        if "merged_entry_tree" in resource_config:
            self.resource.update(
                merged_entry_tree=EntryTree().fromdict(
                    resource_config["merged_entry_tree"]
                ),
            )

    def _dump_to_config(self):
        r"""Serialize the pipeline to an IR(intermediate representation).
        The returned IR can be passed to `init_from_config` to initialize
        a pipeline.

        Returns:
            dict: A dictionary storing IR.
        """

        def test_jsonable(test_dict: Dict, err_msg: str):
            r"""Check if a dictionary is JSON serializable"""
            try:
                json.dumps(test_dict)
                return test_dict
            except (TypeError, OverflowError) as e:
                raise ProcessorConfigError(err_msg) from e

        get_err_msg: Dict = {
            "reader": lambda reader: (
                "The reader of the pipeline cannot be JSON serialized. This is"
                " likely due to some parameters in the configuration of the "
                f"reader {get_full_module_name(reader)} cannot be serialized "
                "in JSON. To resolve this issue, you can consider implementing"
                " a JSON serialization for that parameter type or changing the"
                " parameters of this reader. Note that in order for the reader"
                " to be serialized in JSON, all the variables defined in both "
                "the default_configs and the configuration passed in during "
                "pipeline.set_reader() need to be JSON-serializable. You can "
                "find in the stack trace the type of the un-serializable "
                "parameter."
            ),
            "component": lambda component: (
                "One component of the pipeline cannot be JSON serialized. This"
                " is likely due to some parameters in the configuration of the"
                f" component {get_full_module_name(component)} cannot be "
                "serialized in JSON. To resolve this issue, you can consider "
                "implementing a JSON serialization for that parameter type or "
                "changing the parameters of the component. Note that in order "
                "for the component to be serialized in JSON, all the variables"
                " defined in both the default_configs and the configuration "
                "passed in during pipeline.add() need to be JSON-serializable."
                " You can find in the stack trace the type of the "
                "un-serializable parameter."
            ),
            "selector": lambda selector: (
                "A selector cannot be JSON serialized. This is likely due to "
                "some __init__ parameters for class "
                f"{get_full_module_name(selector)} cannot be serialized in "
                "JSON. To resolve this issue, you can consider implementing a "
                "JSON serialization for that parameter type or changing the "
                "signature of the __init__ function. You can find in the stack"
                " trace the type of the un-serializable parameter."
            ),
        }

        configs: Dict = {
            "forte_ir_version": FORTE_IR_VERSION,
            "components": [],
            "states": {},
        }

        # Serialize pipeline components
        configs["components"].append(
            {
                "type": get_full_module_name(self._reader),
                "configs": test_jsonable(
                    test_dict=self._reader_config.todict(),
                    err_msg=get_err_msg["reader"](self._reader),
                ),
            }
        )
        for component, config, selector, selector_config in zip(
            self.components,
            self.component_configs,
            self._selectors,
            self._selectors_configs,
        ):
            configs["components"].append(
                {
                    "type": get_full_module_name(component),
                    "configs": test_jsonable(
                        test_dict=config.todict(),
                        err_msg=get_err_msg["component"](component),
                    ),
                    "selector": {
                        "type": get_full_module_name(selector),
                        "configs": test_jsonable(
                            # pylint: disable=protected-access
                            test_dict=selector_config.todict(),
                            # pylint: enable=protected-access
                            err_msg=get_err_msg["selector"](selector),
                        ),
                    },
                }
            )

        # Serialize current states of pipeline
        configs["states"].update(
            {
                "attribute": {
                    attr: getattr(self, attr)
                    for attr in (
                        "_initialized",
                        "_enable_profiling",
                        "_check_type_consistency",
                        "_do_init_type_check",
                    )
                    if hasattr(self, attr)
                },
                "resource": {},
            }
        )
        if self.resource.contains("onto_specs_dict"):
            configs["states"]["resource"].update(
                {"onto_specs_dict": self.resource.get("onto_specs_dict")}
            )
        if self.resource.contains("merged_entry_tree"):
            configs["states"]["resource"].update(
                {
                    "merged_entry_tree": self.resource.get(
                        "merged_entry_tree"
                    ).todict()
                }
            )

        return configs

    def save(self, path: str):
        r"""Store the pipeline as an IR(intermediate representation) in yaml.
        The path can then be passed to
        :meth:`~forte.pipeline.Pipeline.init_from_config_path` to initialize
        a pipeline. Note that calling
        :meth:`~forte.pipeline.Pipeline.init_from_config` from a different
        python environment may not work for some self defined component classes
        because their module name is `__main__`.

        Args:
            path: The file path to save configurations.
        """
        with open(path, "w", encoding="utf-8") as f:
            yaml.safe_dump(self._dump_to_config(), f)

    def export(self, name: Optional[str] = None) -> Optional[str]:
        r"""Exports pipeline to FORTE_EXPORT_PATH.

        FORTE_EXPORT_PATH is a directory where all serialized pipeline will be stored.
        Users can specify through environment variable FORTE_EXPORT_PATH.

        This method will have the following behaviors:

            - FORTE_EXPORT_PATH will be created if assigned but not found.

            - If name is not provided, a default name `pipeline` will be used
              and suffixed by UUID, to prevent overwriting
              (e.g. `pipeline-4ba29336-aa05-11ec-abec-309c23414763.yml`).

            - If name is provided, then no suffix will be appended.

            - The pipeline is saved by :meth:`~forte.pipeline.Pipeline.save`,
              which exports the pipeline by :meth:`~forte.pipeline.Pipeline._dump_to_config`
              and saves it to a YAML file.

        Args:
            name: Export name of the pipeline. Default is None.

        Returns:
            Optional[str]: Export path of pipeline config YAML.

        Raises:
            ValueError: if export name is already taken.
        """
        export_path: Optional[str] = os.environ.get("FORTE_EXPORT_PATH")
        if export_path:
            os.makedirs(export_path, exist_ok=True)

            if name:
                export_name = f"{name}.yml"
            else:
                suffix = str(uuid.uuid1())
                export_name = f"pipeline-{suffix}.yml"

            export_path = os.path.join(export_path, export_name)
            if Path(export_path).exists():
                raise ValueError(f"{export_path} already exists.")

            self.save(export_path)
        return export_path

    def _remote_service_app(
        self, service_name: str = "", input_format: str = "string"
    ):
        r"""Return a FastAPI app that can be used to serve the pipeline.

        Args:
            service_name: Assign a name to the pipeline service for validation.
                This will appear in the `service_name` field on default page
                and can be queried and validated against the expected service
                name set by user. Default to `''`.
            input_format: Specify format of the input for validation. It can be
                `"string"` or `"DataPack"`. This will appear in the
                `input_format` field on default page and can be queried and
                validated against the expected input format set by user.
                Default to `"string"`.

        Returns:
            FastAPI: A FastAPI app for remote service.

        Raises:
            ImportError: An error occurred importing `fastapi` module.
        """
        # TODO: Currently we only support the `process` function, but it can
        # be extended by adding new interfaces that wrap up any Pipeline
        # method. Refer to https://fastapi.tiangolo.com for more info.
        try:
            # pylint:disable=import-outside-toplevel
            from fastapi import FastAPI

            # pydantic should be installed as dependency of fastapi
            from pydantic import BaseModel
        except ImportError as e:
            raise ImportError(
                create_import_error_msg(
                    "fastapi", "remote", "the pipeline service"
                )
            ) from e

        app = FastAPI()
        records: Optional[Dict[str, Set[str]]] = None

        class RequestBody(BaseModel):
            args: str = "[]"
            kwargs: str = "{}"

        # pylint: disable=unused-variable
        @app.get("/")
        def default_page():
            return {
                "status": "OK",
                "service_name": service_name,
                "input_format": input_format,
                "pipeline": self._dump_to_config(),
            }

        @app.get("/records")
        def get_records():
            nonlocal records
            if records is None:
                # Collect records of each pipeline component for validation
                records = {}
                for component in [self._reader] + self.components:
                    if hasattr(component, "record"):
                        component.record(records)
            return {"status": "OK", "records": records}

        @app.get("/expectation")
        def get_expectation():
            expectation: Dict[str, Set[str]] = {}
            if len(self.components) > 0 and hasattr(
                self.components[0], "expected_types_and_attributes"
            ):
                expectation = self.components[0].expected_types_and_attributes()
            return {"status": "OK", "expectation": expectation}

        @app.post("/process")
        def run_pipeline(body: RequestBody):
            args = json.loads(body.args)
            kwargs = json.loads(body.kwargs)
            result = self.process(*args, **kwargs)
            return {"status": "OK", "result": result.to_string()}

        # pylint: enable=unused-variable

        return app

    def serve(
        self,
        host: str = "localhost",
        port: int = 8008,
        service_name: str = "",
        input_format: str = "string",
    ):
        r"""Start a service of the current pipeline at a specified host
        and port.

        Args:
            host: Port number of pipeline service.
            port: Host name of pipeline service.
            service_name: Assign a name to the pipeline service for validation.
                This will appear in the `service_name` field on default page
                and can be queried and validated against the expected service
                name set by user. Default to `''`.
            input_format: Specify format of the input for validation. It can be
                `"string"` or `"DataPack"`. This will appear in the
                `input_format` field on default page and can be queried and
                validated against the expected input format set by user.
                Default to `"string"`.

        Raises:
            ImportError: An error occurred importing `uvicorn` module.
        """
        if "uvicorn" not in sys.modules:
            try:
                import uvicorn  # pylint: disable=import-outside-toplevel
            except ImportError as e:
                raise ImportError(
                    create_import_error_msg(
                        "uvicorn", "remote", "the pipeline service"
                    )
                ) from e

        self.initialize()
        uvicorn.run(
            self._remote_service_app(
                service_name=service_name, input_format=input_format
            ),
            host=host,
            port=port,
            log_level="info",
        )

    def set_profiling(self, enable_profiling: bool = True):
        r"""Set profiling option.

        Args:
            enable_profiling: A boolean of whether to enable profiling
                for the pipeline or not (the default is True).
        """
        self._enable_profiling = enable_profiling

    def initialize(self) -> "Pipeline":
        """
        This function should be called before the pipeline can be used to
        process the actual data. This function will call the `initialize` of
        all the components inside this pipeline.

        Returns:
            None
        """
        # create EntryTree type object merged_entry_tree to store the parsed
        # entry tree from ontology specification file passed in as part of
        # resource and add the result to resource with key of merged_entry_tree.
        merged_entry_tree = EntryTree()
        if self.resource.get("onto_specs_path"):
            OntologyCodeGenerator().parse_schema_for_no_import_onto_specs_file(
                ontology_path=self.resource.get("onto_specs_path"),
                ontology_dict=self.resource.get("onto_specs_dict"),
                merged_entry_tree=merged_entry_tree,
            )
            self.resource.update(merged_entry_tree=merged_entry_tree)

        # The process manager need to be assigned first.
        self._proc_mgr = ProcessManager(len(self._components))

        if self._initialized:
            # The pipeline has already been initialized, so we are doing
            # re-initialization here.
            logging.info("Re-initializing the Pipeline.")

        # Reset the flags of the components before initializing them.
        self._reader.reset_flags()
        for c in self._components:
            c.reset_flags()

        # Handle the reader.
        if not self._reader.is_initialized:
            self._reader.initialize(self.resource, self._reader_config)
        else:
            logging.info(
                "The reader [%s] has already initialized, "
                "will skip its initialization.",
                self._reader.name,
            )

        if self._check_type_consistency:
            self.reader.enforce_consistency(enforce=True)
        else:
            self.reader.enforce_consistency(enforce=False)

        # Handle other components and their selectors.
        self.initialize_components()
        self.initialize_selectors()
        self._initialized = True

        # Create profiler
        if self._enable_profiling:
            self.reader.set_profiling(True)
            self._profiler = [0.0] * len(self.components)

        # Check record types and attributes of each pipeline component
        if self._do_init_type_check:
            current_records: Dict[str, Set[str]] = {}
            self._reader.record(current_records)
            for component in self.components:
                if hasattr(component, "expected_types_and_attributes"):
                    record_types_and_attributes_check(
                        component.expected_types_and_attributes(),  # type: ignore
                        current_records,
                    )
                if hasattr(component, "record"):
                    component.record(current_records)  # type: ignore

        return self

    def initialize_components(self):
        """
        This function will initialize all the components in this pipeline,
        except the reader. The components are initialized in a FIFO manner
        based on the order of insertion,

        During initialization, the component will be configured based on its
        corresponding configuration. However, if the component is already
        initialized (for example, being initialized manually or used twice
        in the same pipeline), the new configuration will be ignored.

        The pipeline will check for type dependencies between the components
        inside this pipeline, see
        :func:`~forte.pipeline_component.PipelineComponent.enforce_consistency`
        for more details.

        """
        for component, config in zip(self.components, self.component_configs):
            try:
                if not component.is_initialized:
                    component.initialize(self.resource, config)
                else:
                    logging.info(
                        "The component [%s] has already initialized, "
                        "will skip its initialization.",
                        component.name,
                    )
            except ProcessorConfigError as e:
                logging.error(
                    "Exception occur when initializing " "processor %s",
                    component.name,
                )
                raise e

            component.enforce_consistency(enforce=self._check_type_consistency)

    def initialize_selectors(self):
        """
        This function will reset the states of selectors
        """
        for selector, config in zip(self._selectors, self._selectors_configs):
            try:
                selector.initialize(config)
            except ValueError as e:
                logging.error("Exception occur when initializing selectors")
                raise e

    def set_reader(
        self,
        reader: BaseReader,
        config: Optional[Union[Config, Dict[str, Any]]] = None,
    ) -> "Pipeline":
        """
        Set the reader of the pipeline. A reader is the entry point of
        this pipeline, data flown into the reader will be converted to the
        data pack format, and being passed onto the other components for
        processing.

        Args:
            reader: The reader to be used of the pipeline
            config: The custom configuration to be passed to the reader. If
              the config is not provided, the default config defined by the
              reader class will be used.

        Returns:
            The pipeline itself, which allows you to directly chain other
            pipeline construction code afterwards, i.e., you can do:

            .. code-block:: python

                Pipeline().set_reader(your_reader()).add(your_processor())

        """
        self._reader = reader
        self._reader_config = reader.make_configs(config)
        return self

    @property
    def reader(self) -> BaseReader:
        return self._reader

    @property
    def components(self) -> List[PipelineComponent]:
        """
        Return all the components in this pipeline, except the reader.

        Returns: A list containing the components.

        """
        return self._components

    @property
    def ref_names(self) -> Dict[str, int]:
        """
        Return all the reference names in this pipeline, except the reader.

        Returns: A dictionary containing the reference names.

        """
        return self._ref_names

    @property
    def component_configs(self) -> List[Optional[Config]]:
        """
        Return the configs related to the components, except the reader.

        Returns: A list containing the components configs.

        """
        return self._configs

    def add(
        self,
        component: PipelineComponent,
        config: Optional[Union[Config, Dict[str, Any]]] = None,
        selector: Optional[Selector] = None,
        selector_config: Optional[Union[Config, Dict[str, Any]]] = None,
        ref_name: Optional[str] = None,
    ) -> "Pipeline":
        """
        Adds a pipeline component to the pipeline. The pipeline components
        will form a chain based on the insertion order. The customized
        `config` and `selector` (:class:`~forte.data.selector.Selector`)
        will be associated with this particular component. If the `config`
        or the `selector` is not provided, the default ones will be used.

        Here, note that the same component instance can be added multiple
        times to the pipeline. In such cases, the instance will only be
        setup at the first insertion (i.e. its `initialize` function will
        only be called once). The subsequent insertion of the same component
        instance will not change the behavior nor the states of the instance.
        Thus, a different `config` cannot be provided (should be `None`) when
        added the second time, otherwise a `ProcessorConfigError` will be
        thrown. In the case where one want to them to behave differently, a
        different instance should be used.

        Args:
            component: The component to be inserted next
              to the pipeline.
            config: The custom configuration
              to be used for the added component. Default None, which means
              the `default_configs()` of the component will be used.
            selector: The selector used to pick the corresponding
              data pack to be consumed by the component. Default None, which
              means the whole pack will be used.

        Returns:
            The pipeline itself, which enables one to chain the creation of
            the pipeline, i.e., you can do:

            .. code-block:: python

                Pipeline().set_reader(your_reader()).add(
                    your_processor()).add(anther_processor())
        """
        if isinstance(component, BaseReader):
            raise ProcessFlowException("Reader need to be set via set_reader()")

        if isinstance(component, Evaluator):
            # This will ask the job to keep a copy of the gold standard.
            self.evaluator_indices.append(len(self.components))

        if ref_name is not None:
            if ref_name in self._ref_names:
                raise ValidationError(
                    f"This reference name {ref_name} already exists, please specify a new one"
                )
            else:
                self._ref_names[ref_name] = len(self.components)

        if component not in self.__component_set:
            # The case where the component is not found.
            self._components.append(component)
            self.__component_set.add(component)
            self.component_configs.append(component.make_configs(config))
        else:
            if config is None:
                self._components.append(component)
                # We insert a `None` value here just to make the config list
                # to match the component list, but this config should not be
                # used.
                self.component_configs.append(None)
            else:
                raise ProcessorConfigError(
                    f"The same instance of a component named {component.name} "
                    f" has already been added to"
                    f" the pipeline, we do not accept a different configuration"
                    f" for it. If you would like to use a differently"
                    f" configured component, please create another instance."
                    f" If you intend to re-use the component instance, please"
                    f" do not provide the `config` (or provide a `None`)."
                )

        if selector is None:
            self._selectors.append(self.__default_selector)
            self._selectors_configs.append(self.__default_selector_config)
        else:
            self._selectors.append(selector)
            self._selectors_configs.append(
                selector.make_configs(selector_config)
            )

        return self

    def add_gold_packs(self, pack):
        r"""Add gold packs to a internal dictionary used for evaluation.
        This dictionary is used by the evaluator while calling
        `consume_next(...)`

        Args:
            pack: A key, value pair containing job.id -> gold_pack
                mapping
        """
        self._predict_to_gold.update(pack)

    def process(self, *args, **kwargs) -> PackType:
        r"""Alias for :meth:`process_one`.

        Args:
            args: The positional arguments used to get the initial data.
            kwargs: The keyword arguments used to get the initial data.
        """
        return self.process_one(*args, **kwargs)

    def run(self, *args, **kwargs):
        r"""Run the whole pipeline and ignore all returned DataPack. This is
        mostly used when you need to run the pipeline and do not require the
        output but rely on the side-effect. For example, if the pipeline
        writes some data to disk.

        Calling this function will automatically call the :meth:`initialize`
        at the beginning, and call the :meth:`finish` at the end.

        Args:
            args: The positional arguments used to get the initial data.
            kwargs: The keyword arguments used to get the initial data.
        """
        self.initialize()
        for _ in self.process_dataset(*args, **kwargs):
            # Process the whole dataset ignoring the return values.
            # This essentially expect the processors have side effects.
            pass
        self.finish()

    def process_one(self, *args, **kwargs) -> PackType:
        r"""Process one single data pack. This is done by only reading and
        processing the first pack in the reader.

        Args:
            kwargs: the information needed to load the data. For example, if
                :attr:`_reader` is
                :class:`~forte.data.readers.string_reader.StringReader`, this
                should contain a
                single piece of text in the form of a string variable. If
                :attr:`_reader` is a file reader, this can point to the file
                path.
        """
        if not self._initialized:
            raise ProcessFlowException(
                "Please call initialize before running the pipeline"
            )

        first_pack = []

        for p in self._reader.iter(*args, **kwargs):
            first_pack.append(p)
            break

        if len(first_pack) == 1:
            results = list(self._process_packs(iter(first_pack)))
            return results[0]
        else:
            raise ValueError("Input data source contains no packs.")

    def process_dataset(self, *args, **kwargs) -> Iterator[PackType]:
        r"""Process the documents in the data source(s) and return an
        iterator or list of DataPacks. The arguments are directly passed
        to the reader to take data from the source.
        """
        if not self._initialized:
            raise ProcessFlowException(
                "Please call initialize before running the pipeline"
            )

        data_iter = self._reader.iter(*args, **kwargs)
        return self._process_packs(data_iter)

    def finish(self):
        """
        Call the finish method of all pipeline component. This need to be called
        explicitly to release all resources.
        """

        # Report time profiling of readers and processors
        if self._enable_profiling:
            out_header: str = "Pipeline Time Profile\n"
            out_reader: str = (
                f"- Reader: {self.reader.component_name}, "
                + f"{self.reader.time_profile} s\n"
            )
            out_processor: str = "\n".join(
                [
                    f"- Component [{i}]: {self.components[i].name}, {t} s"
                    for i, t in enumerate(self._profiler)
                ]
            )
            logger.info("%s%s%s", out_header, out_reader, out_processor)

        self.reader.finish(self.resource)
        for p in self.components:
            p.finish(self.resource)
        self._initialized = False

    def __update_stream_job_status(self):
        q_index = self._proc_mgr.current_queue_index
        u_index = self._proc_mgr.unprocessed_queue_indices[q_index]
        current_queue = self._proc_mgr.current_queue

        for job_i in itertools.islice(current_queue, 0, u_index + 1):
            if job_i.status == ProcessJobStatus.UNPROCESSED:
                job_i.set_status(ProcessJobStatus.PROCESSED)

    def __update_batch_job_status(self, component: BaseBatchProcessor):
        # update the status of the jobs. The jobs which were removed from
        # data_pack_pool will have status "PROCESSED" else they are "QUEUED"
        q_index = self._proc_mgr.current_queue_index
        u_index = self._proc_mgr.unprocessed_queue_indices[q_index]
        current_queue = self._proc_mgr.current_queue

        data_pool_length = len(component.batcher.data_pack_pool)

        for i, job_i in enumerate(
            itertools.islice(current_queue, 0, u_index + 1)
        ):
            if i <= u_index - data_pool_length:
                job_i.set_status(ProcessJobStatus.PROCESSED)
            else:
                job_i.set_status(ProcessJobStatus.QUEUED)

    def __flush_batch_job_status(self):
        current_queue = self._proc_mgr.current_queue
        for job in current_queue:
            job.set_status(ProcessJobStatus.PROCESSED)

    def _process_with_component(
        self,
        selector: Selector,
        component: PipelineComponent,
        raw_job: ProcessJob,
    ):
        for pack in selector.select(raw_job.pack):
            # First, perform the component action on the pack
            try:
                if isinstance(component, Caster):
                    # Replacing the job pack with the casted version.
                    raw_job.alter_pack(component.cast(pack))
                elif isinstance(component, BaseBatchProcessor):
                    pack.set_control_component(component.name)
                    component.process(pack)
                elif isinstance(component, Evaluator):
                    pack.set_control_component(component.name)
                    component.consume_next(
                        pack, self._predict_to_gold[raw_job.id]
                    )
                elif isinstance(component, BaseProcessor):
                    # Should be BasePackProcessor:
                    # All other processor are considered to be
                    # streaming processor like this.
                    pack.set_control_component(component.name)
                    component.process(pack)
                # After the component action, make sure the entry is
                # added into the index.
                pack.add_all_remaining_entries()
            except ValueError as e:
                logger.error(e, exc_info=True)
                raise ProcessExecutionException(
                    f"Exception occurred when running " f"{component.name}"
                ) from e

    def _process_packs(
        self, data_iter: Iterator[PackType]
    ) -> Iterator[PackType]:
        r"""Process the packs received from the reader by the running through
        the pipeline.

        Args:
             data_iter: Iterator yielding jobs that contain packs

        Returns:
            Yields packs that are processed by the pipeline.
        """

        # pylint: disable=line-too-long

        # Here is the logic for the execution of the pipeline.

        # The basic idea is to yield a pack as soon as it gets processed by all
        # the processors instead of waiting for later jobs to get processed.

        # 1) A job can be in three status
        #  - UNPROCESSED
        #  - QUEUED
        #  - PROCESSED
        #
        # 2) Each processor maintains a queue to hold the jobs to be executed
        # next.
        #
        # 3) In case of a BatchProcessor, a job enters into QUEUED status if the
        # batch is not full according to the batcher of that processor.
        # In that case, the pipeline requests for additional jobs from the
        # reader and starts the execution loop from the beginning.
        #
        # 4) At any point, while moving to the next processor, the pipeline
        # ensures that all jobs are either in QUEUED or PROCESSED status. If
        # they are PROCESSED, they will be moved to the next queue. This design
        # ensures that at any point, while processing the job at processor `i`,
        # all the jobs in the previous queues are in QUEUED status. So whenever
        # a new job is needed, the pipeline can directly request it from the
        # reader instead of looking at previous queues for UNPROCESSED jobs.
        #
        # 5) When a processor receives a poison pack, it flushes all the
        # remaining batches in its memory (this actually has no effect in
        # PackProcessors) and moves the jobs including the poison pack to the
        # next queue. If there is no next processor, the packs are yield.
        #
        # 6) The loop terminates when the last queue contains only a poison pack
        #
        # Here is the sample pipeline and its execution
        #
        # Assume 1 pack corresponds to a batch of size 1
        #
        # After 1st step (iteration), reading from the reader,
        #
        #            batch_size = 2                               batch_size = 2
        #  Reader -> B1 (BatchProcessor) -> P1 (PackProcessor) -> B2(BatchProcessor)
        #
        #          |___________|
        #          |___________|
        #          |___________|
        #          |___________|
        #          |_J1:QUEUED_|
        #
        # B1 needs another pack to process job J1
        #
        # After 2nd step (iteration),
        #
        #           batch_size = 2                               batch_size = 2
        # Reader -> B1 (BatchProcessor) -> P1 (PackProcessor) -> B2(BatchProcessor)
        #
        #          |___________|       |_______________|
        #          |___________|       |_______________|
        #          |___________|       |_______________|
        #          |___________|       |_J2:UNPROCESSED_|
        #          |___________|       |_J1:UNPROCESSED_|
        #
        # B1 processes both the packs, the jobs are moved to the next queue.
        #
        # After 3rd step (iteration),
        #
        #           batch_size = 2                               batch_size = 2
        # Reader -> B1 (BatchProcessor) -> P1 (PackProcessor) -> B2(BatchProcessor)
        #
        #          |___________|       |_______________|     |_______________|
        #          |___________|       |_______________|     |_______________|
        #          |___________|       |_______________|     |_______________|
        #          |___________|       |_______________|     |_______________|
        #          |___________|       |_J2:UNPROCESSED_|     |_J1:UNPROCESSED_|
        #
        # P1 processes the first job. However, there exists one UNPROCESSED job
        # J2 in the queue. Pipeline first processes this job before moving to the
        # next processor
        #
        # After 4th step (iteration),
        #
        #           batch_size = 2                               batch_size = 2
        # Reader -> B1 (BatchProcessor) -> P1 (PackProcessor) -> B2(BatchProcessor)
        #
        #        |___________|       |_______________|     |_______________|
        #        |___________|       |_______________|     |_______________|
        #        |___________|       |_______________|     |_______________|
        #        |___________|       |_______________|     |_J2:UNPROCESSED_|
        #        |___________|       |_______________|     |_J1:UNPROCESSED_|
        #
        #
        # After 5th step (iteration),
        #
        #           batch_size = 2                               batch_size = 2
        # Reader -> B1 (BatchProcessor) -> P1 (PackProcessor) -> B2(BatchProcessor)
        #
        #        |___________|       |_______________|     |_______________|
        #        |___________|       |_______________|     |_______________|
        #        |___________|       |_______________|     |_______________|    --> Yield J1.pack and J2.pack
        #        |___________|       |_______________|     |_______________|
        #        |___________|       |_______________|     |_______________|

        if not self._initialized:
            raise ProcessFlowException(
                "Please call initialize before running the pipeline"
            )

        buffer = ProcessBuffer(self, data_iter)

        if len(self.components) == 0:
            yield from data_iter
            # Write return here instead of using if..else to reduce indent.
            return

        while not self._proc_mgr.exhausted():
            # Take the raw job from the buffer, the job status now should
            # be UNPROCESSED.
            raw_job: ProcessJob = next(buffer)

            component_index = self._proc_mgr.current_processor_index
            component = self.components[component_index]
            selector: Selector = self._selectors[component_index]
            current_queue_index = self._proc_mgr.current_queue_index
            current_queue: Deque[ProcessJob] = self._proc_mgr.current_queue
            pipeline_length = self._proc_mgr.pipeline_length
            unprocessed_queue_indices = self._proc_mgr.unprocessed_queue_indices
            processed_queue_indices = self._proc_mgr.processed_queue_indices
            next_queue_index = current_queue_index + 1
            should_yield = next_queue_index >= pipeline_length

            if not raw_job.is_poison:
                # Start timer
                if self._enable_profiling:
                    start_time: float = time()

                self._process_with_component(selector, component, raw_job)

                # Stop timer and add to time profiler
                if self._enable_profiling:
                    self._profiler[component_index] += time() - start_time

                # Then, based on component type, handle the queue.
                if isinstance(component, BaseBatchProcessor):
                    self.__update_batch_job_status(component)
                    index = unprocessed_queue_indices[current_queue_index]

                    # Check status of all the jobs up to "index".
                    for i, job_i in enumerate(
                        itertools.islice(current_queue, 0, index + 1)
                    ):
                        if job_i.status == ProcessJobStatus.PROCESSED:
                            processed_queue_indices[current_queue_index] = i

                    # There are UNPROCESSED jobs in the queue.
                    if index < len(current_queue) - 1:
                        unprocessed_queue_indices[current_queue_index] += 1
                    elif processed_queue_indices[current_queue_index] == -1:
                        # Fetch more data from the reader to process the
                        # first job.
                        unprocessed_queue_indices[current_queue_index] = len(
                            current_queue
                        )
                        self._proc_mgr.current_processor_index = 0
                        self._proc_mgr.current_queue_index = -1
                    else:
                        processed_queue_index = processed_queue_indices[
                            current_queue_index
                        ]
                        # Move or yield the pack.
                        c_queue = list(current_queue)
                        for job_i in c_queue[: processed_queue_index + 1]:
                            if job_i.status == ProcessJobStatus.PROCESSED:
                                if should_yield:
                                    if job_i.id in self._predict_to_gold:
                                        self._predict_to_gold.pop(job_i.id)
                                    # TODO: I don't know why these are
                                    #  marked as incompatible type by mypy.
                                    #  the same happens 3 times on every yield.
                                    #  It is observed that the pack returned
                                    #  from the `ProcessJob` is considered to
                                    #  be different from `PackType`.
                                    yield job_i.pack  # type: ignore
                                else:
                                    self._proc_mgr.add_to_queue(
                                        queue_index=next_queue_index, job=job_i
                                    )
                            else:
                                raise ProcessFlowException(
                                    f"The job status should be "
                                    f"{ProcessJobStatus.PROCESSED} "
                                    f"at this point."
                                )
                            current_queue.popleft()

                        # Set the UNPROCESSED and PROCESSED indices.
                        unprocessed_queue_indices[current_queue_index] = len(
                            current_queue
                        )

                        processed_queue_indices[current_queue_index] = -1

                        if should_yield:
                            self._proc_mgr.current_processor_index = 0
                            self._proc_mgr.current_queue_index = -1
                        else:
                            self._proc_mgr.current_processor_index = (
                                next_queue_index
                            )
                            self._proc_mgr.current_queue_index = (
                                next_queue_index
                            )
                # Besides Batch Processors, the other component type only
                # deal with one pack at a time, these include: PackProcessor
                # Evaluator, Caster.
                # - Move them to the next queue
                else:
                    self.__update_stream_job_status()
                    index = unprocessed_queue_indices[current_queue_index]

                    # there are UNPROCESSED jobs in the queue
                    if index < len(current_queue) - 1:
                        unprocessed_queue_indices[current_queue_index] += 1
                    else:
                        # current_queue is modified in this array
                        for job_i in list(current_queue):
                            if job_i.status == ProcessJobStatus.PROCESSED:
                                if should_yield:
                                    if job_i.id in self._predict_to_gold:
                                        self._predict_to_gold.pop(job_i.id)
                                    yield job_i.pack  # type: ignore
                                else:
                                    self._proc_mgr.add_to_queue(
                                        queue_index=next_queue_index, job=job_i
                                    )
                                current_queue.popleft()
                            else:
                                raise ProcessFlowException(
                                    f"The job status should be "
                                    f"{ProcessJobStatus.PROCESSED} "
                                    f"at this point."
                                )

                        # set the UNPROCESSED index
                        # we do not use "processed_queue_indices" as the
                        # jobs get PROCESSED whenever they are passed
                        # into a PackProcessor
                        unprocessed_queue_indices[current_queue_index] = len(
                            current_queue
                        )

                        # update the current queue and processor only
                        # when all the jobs are processed in the current
                        # queue
                        if should_yield:
                            self._proc_mgr.current_processor_index = 0
                            self._proc_mgr.current_queue_index = -1

                        else:
                            self._proc_mgr.current_processor_index = (
                                next_queue_index
                            )
                            self._proc_mgr.current_queue_index = (
                                next_queue_index
                            )
            else:
                component.flush()
                self.__flush_batch_job_status()

                # current queue is modified in the loop
                for job in list(current_queue):
                    if (
                        job.status != ProcessJobStatus.PROCESSED
                        and not job.is_poison
                    ):
                        raise ValueError(
                            "Job is neither PROCESSED nor is "
                            "a poison. Something went wrong "
                            "during execution."
                        )

                    if not job.is_poison and should_yield:
                        if job.id in self._predict_to_gold:
                            self._predict_to_gold.pop(job.id)
                        yield job.pack  # type: ignore

                    elif not should_yield:
                        self._proc_mgr.add_to_queue(
                            queue_index=next_queue_index, job=job
                        )

                    if not job.is_poison:
                        current_queue.popleft()

                if not should_yield:
                    # set next processor and queue as current
                    self._proc_mgr.current_processor_index = next_queue_index
                    self._proc_mgr.current_queue_index = next_queue_index

        self._proc_mgr.reset()

    def evaluate(self) -> Iterator[Tuple[str, Any]]:
        """
        Call the evaluators in the pipeline to collect their results.

        Returns:
            Iterator of the evaluator results. Each element is a tuple, where
            the first one is the name of the evaluator, and the second one
            is the output of the evaluator (see
            :func:`~forte.evaluation.base.base_evaluator.Evaluator.get_result`).
        """
        for i in self.evaluator_indices:
            p = self.components[i]
            assert isinstance(p, Evaluator)
            yield p.name, p.get_result()

    def get_component(self, ref_name: str) -> PipelineComponent[Any]:
        """
        Call the evaluator in the pipeline by the reference name to get a component.

        Args:
            ref_name: the reference name of a component
        """
        p = self.components[self.ref_names[ref_name]]
        return p


def serve(
    pl_config_path: str,
    host: str = "localhost",
    port: int = 8008,
    service_name: str = "",
    input_format: str = "string",
):
    r"""Start a remote service of a pipeline initialized from a YAML config at
    a specified host and port.

    Args:
        pl_config_path: A string of the configuration path, which is
            is a YAML file that specify the structure and parameters of the
            pipeline.
        host: Port number of pipeline service.
        port: Host name of pipeline service.
        service_name: Assign a name to the pipeline service for validation.
                This will appear in the `service_name` field on default page
                and can be queried and validated against the expected service
                name set by user. Default to `''`.
        input_format: Specify format of the input for validation. It can be
            `"string"` or `"DataPack"`. This will appear in the
            `input_format` field on default page and can be queried and
            validated against the expected input format set by user.
            Default to `"string"`.
    """
    pipeline: Pipeline = Pipeline()
    pipeline.init_from_config_path(pl_config_path)
    pipeline.serve(
        host=host,
        port=port,
        service_name=service_name,
        input_format=input_format,
    )



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads plain text data into Datapacks.
"""
import logging
from typing import Iterator, List, Union, Dict, Set

from forte.data.data_pack import DataPack
from forte.data.base_reader import PackReader
from ft.onto.base_ontology import Document

logger = logging.getLogger(__name__)

__all__ = [
    "StringReader",
]


class StringReader(PackReader):
    r"""
    :class:`StringReader` is designed to read in a list of string variables.
    """

    # pylint: disable=unused-argument
    def _cache_key_function(self, collection) -> str:
        return str(hash(collection)) + ".html"

    def _collect(  # type: ignore
        self, string_data: Union[List[str], str]
    ) -> Iterator[str]:
        r"""``string`_data` should be of type `List[str]`,
        which is the list of raw text strings to iterate over.
        """
        # This allows the user to pass in either one single string or a list of
        # strings.
        data_strings = (
            [string_data] if isinstance(string_data, str) else string_data
        )
        for data in data_strings:
            yield data

    def _parse_pack(self, data_source: str) -> Iterator[DataPack]:
        r"""Takes a raw string and converts into a :class:`~forte.data.data_pack.DataPack`.

        Args:
            data_source: str that contains text of a document.

        Returns: :class:`~forte.data.data_pack.DataPack` containing Document.
        """
        pack = DataPack()

        self.set_text(pack, data_source)
        Document(pack, 0, len(data_source))

        yield pack

    def record(self, record_meta: Dict[str, Set[str]]):
        r"""Method to add output type record of `StringReader` which is
        `ft.onto.base_ontology.Document` with an empty set
        to :attr:`forte.data.data_pack.Meta.record`.

        Args:
            record_meta: the field in the datapack for type record that need to
                fill in for consistency checking.
        """
        record_meta["ft.onto.base_ontology.Document"] = set()



# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Writers are simply processors with the side-effect to write to the disk.
This file provide some basic writer implementations.
"""
import logging
import os
import posixpath
from abc import abstractmethod, ABC
from typing import Optional, Any, Dict

from forte.common import ProcessorConfigError
from forte.common.configuration import Config
from forte.common.resources import Resources
from forte.data.base_pack import BasePack
from forte.data.data_pack import DataPack
from forte.data.multi_pack import MultiPack
from forte.processors.base.pack_processor import (
    PackProcessor,
    MultiPackProcessor,
)
from forte.utils.utils_io import maybe_create_dir, ensure_dir

logger = logging.getLogger(__name__)

__all__ = [
    "PackWriter",
    "MultiPackWriter",
]


def write_pack(
    input_pack: BasePack,
    output_dir: str,
    sub_path: str,
    indent: Optional[int] = None,
    zip_pack: bool = False,
    overwrite: bool = False,
    drop_record: bool = False,
    serialize_method: str = "jsonpickle",
) -> str:
    """
    Write a pack to a path.

    Args:
        input_pack: A Pack to be written.
        output_dir: The output directory.
        sub_path: The file name for this pack.
        indent: Whether to format JSON with an indent.
        zip_pack: Whether to zip the output JSON.
        overwrite: Whether to overwrite the file if already exists.
        drop_record: Whether to drop the creation records in the serialization.
        serialize_method: The method used to serialize the data. Current
          available options are `jsonpickle` and `pickle`.
          Default is `jsonpickle`.

    Returns:
        If successfully written, will return the path of the output file.
        otherwise, will return None.

    """
    output_path = os.path.join(output_dir, sub_path)

    if overwrite or not os.path.exists(output_path):
        ensure_dir(output_path)
        input_pack.serialize(
            output_path,
            zip_pack=zip_pack,
            drop_record=drop_record,
            serialize_method=serialize_method,
            indent=indent,
        )
    else:
        logging.info("Will not overwrite existing path %s", output_path)

    logging.info("Writing a pack to %s", output_path)
    return output_path


class PackWriter(PackProcessor, ABC):
    def __init__(self):
        super().__init__()
        self._zip_pack: bool = False
        self._indent: Optional[int] = None
        self._suffix: str = ""

    def initialize(self, resources: Resources, configs: Config):
        super().initialize(resources, configs)

        if not configs.output_dir:
            raise NotADirectoryError(
                "Root output directory is not defined "
                "correctly in the configs."
            )

        if not os.path.exists(configs.output_dir):
            os.makedirs(configs.output_dir)

        self._zip_pack = configs.zip_pack
        self._indent = configs.indent

        if self.configs.serialize_method == "jsonpickle":
            self._suffix = ".json.gz" if self._zip_pack else ".json"
        else:
            self._suffix = ".pickle.gz" if self._zip_pack else ".pickle"

    @abstractmethod
    def sub_output_path(self, pack: DataPack) -> Optional[str]:
        r"""Allow defining output path using the information of the pack. If
        `None` is returned,

        Args:
            pack: The input datapack.
        """
        raise NotImplementedError

    @classmethod
    def default_configs(cls):
        r"""This defines a basic configuration structure for writer.

        Here:
          - output_dir (str): the directory for writing the result.

          - zip_pack (bool): whether to zip the data pack. The default value is
             False.

          - indent (int): None not indented, if larger than 0, the JSON
             files will be written in the with the provided indention. The
             default value is None.

          - drop_record: whether to drop the creation records in the data pack,
             the default value is False.

          - serialize_method: The method used to serialize the data. Current
              available options are `jsonpickle` and `pickle`. Default is
              "jsonpickle".

        Returns: The default configuration of this writer.
        """
        return {
            "output_dir": None,
            "zip_pack": False,
            "indent": None,
            "drop_record": False,
            "serialize_method": "jsonpickle",
        }

    def _process(self, input_pack: DataPack):
        sub_path = self.sub_output_path(input_pack)
        if sub_path is not None and not sub_path == "":
            # Sub path could be empty, which we will skip writing the file.
            maybe_create_dir(self.configs.output_dir)
            write_pack(
                input_pack,
                self.configs.output_dir,
                sub_path,
                self.configs.indent,
                self.configs.zip_pack,
                self.configs.overwrite,
                self.configs.drop_record,
                self.configs.serialize_method,
            )


class MultiPackWriter(MultiPackProcessor):
    pack_base_out = "packs"
    multi_base = "multi"
    pack_idx = "pack.idx"
    multi_idx = "multi.idx"

    def initialize(self, resources: Resources, configs: Config):
        # pylint: disable=attribute-defined-outside-init,consider-using-with
        super().initialize(resources, configs)

        if self.configs.output_dir is None:
            raise ProcessorConfigError(
                "`output_dir` is not specified for the writer."
            )

        pack_paths = os.path.join(self.configs.output_dir, self.pack_idx)
        ensure_dir(pack_paths)
        self.pack_idx_out = open(pack_paths, "w", encoding="utf-8")

        multi_index = os.path.join(self.configs.output_dir, self.multi_idx)
        ensure_dir(multi_index)
        self.multi_idx_out = open(multi_index, "w", encoding="utf-8")

        if self.configs.serialize_method == "jsonpickle":
            self._suffix = ".json.gz" if self.configs.zip_pack else ".json"
        else:
            self._suffix = ".pickle.gz" if self.configs.zip_pack else ".pickle"

    def multipack_name(self, pack: MultiPack) -> str:
        r"""Allow defining output path using the information of the multi-pack.
        Extending this path allows one to specify the output file name. Default
        value is a name including the `pack_id` of this multi-pack.

        Args:
            pack: The input multi-pack.
        """
        return f"multi_pack_{pack.pack_id}"

    def _process(self, input_pack: MultiPack):
        multi_out_dir = os.path.join(self.configs.output_dir, self.multi_base)
        pack_out_dir = os.path.join(self.configs.output_dir, self.pack_base_out)

        for pack in input_pack.packs:
            pack_out = write_pack(
                pack,
                pack_out_dir,
                str(pack.pack_id) + self._suffix,
                self.configs.indent,
                self.configs.zip_pack,
                self.configs.overwrite,
                self.configs.drop_record,
                self.configs.serialize_method,
            )

            self.pack_idx_out.write(
                f"{pack.pack_id}\t"
                f"{posixpath.relpath(pack_out, self.configs.output_dir)}\n"
            )

        multi_out = write_pack(
            input_pack,
            multi_out_dir,
            self.multipack_name(input_pack) + self._suffix,
            self.configs.indent,
            self.configs.zip_pack,
            self.configs.overwrite,
            self.configs.drop_record,
            self.configs.serialize_method,
        )

        self.multi_idx_out.write(
            f"{input_pack.pack_id}\t"
            f"{posixpath.relpath(multi_out, self.configs.output_dir)}\n"
        )

    def finish(self, _):
        self.pack_idx_out.close()
        self.multi_idx_out.close()

    @classmethod
    def default_configs(cls) -> Dict[str, Any]:
        return {
            "output_dir": None,
            "zip_pack": False,
            "indent": None,
            "drop_record": False,
            "serialize_method": "jsonpickle",
        }



# Copyright 2020 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file implements CharExtractor, which is used to extract feature
from characters of a piece of text.
"""
import logging
from typing import Optional

from forte.data.data_pack import DataPack
from forte.data.converter.feature import Feature
from forte.data.base_extractor import BaseExtractor
from forte.data.ontology import Annotation

logger = logging.getLogger(__name__)

__all__ = ["CharExtractor"]


class CharExtractor(BaseExtractor):
    r"""CharExtractor extracts feature from the text of entry.
    Text will be split into characters.
    """

    @classmethod
    def default_configs(cls):
        r"""Returns a dictionary of default configuration parameters.

        Here:

        - "max_char_length": int
          The maximum number of characters for one token in the text,
          default is None, which means no limit will be set.
        - "entry_type": str
          The fully qualified name of an annotation type entry. Characters
          will be extracted based on these entries. Default is `Token`,
          which means characters of tokens will be extracted.
        """
        config = super().default_configs()
        config.update(
            {
                "max_char_length": None,
                "entry_type": "ft.onto.base_ontology.Token",
            }
        )
        return config

    def update_vocab(
        self, pack: DataPack, context: Optional[Annotation] = None
    ):
        r"""Add all character into vocabulary.

        Args:
            pack: The input data pack.
            context: The context is an Annotation entry where
                features will be extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.
        """
        word: Annotation
        for word in pack.get(self.config.entry_type, context):
            for char in word.text:  # type: ignore
                self.add(char)

    def extract(
        self, pack: DataPack, context: Optional[Annotation] = None
    ) -> Feature:
        r"""Extract the character feature of one instance.

        Args:
            pack: The datapack to extract features from.
            context: The context is an Annotation entry where
                features will be extracted within its range. If None, then the
                whole data pack will be used as the context. Default is None.

        Returns:
            a iterator of feature that contains the characters of each
            specified annotation.
        """
        data = []

        entry: Annotation
        for entry in pack.get(self.config.entry_type, context):
            if self.config.max_char_length is not None:
                max_char_length = min(
                    self.config.max_char_length, len(entry.text)  # type: ignore
                )
            else:
                max_char_length = len(entry.text)  # type: ignore

            characters = entry.text[:max_char_length]  # type: ignore

            if self.vocab:
                data.append([self.element2repr(char) for char in characters])
            else:
                data.append(list(characters))

        meta_data = {
            "need_pad": self.config.need_pad,
            "pad_value": self.get_pad_value(),
            "dim": 2,
            "dtype": int if self.vocab else str,
        }
        return Feature(data=data, metadata=meta_data, vocab=self.vocab)



# Copyright 2022 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The reader that reads audio files into Datapacks.
"""
import os
from typing import Any, Iterator

from forte.data.data_pack import DataPack
from forte.data.data_utils_io import dataset_path_iterator
from forte.data.base_reader import PackReader

__all__ = [
    "AudioReader",
]


class AudioReader(PackReader):
    r""":class:`AudioReader` is designed to read in audio files."""

    def __init__(self):
        super().__init__()
        try:
            import soundfile  # pylint: disable=import-outside-toplevel
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "AudioReader requires 'soundfile' package to be installed."
                " You can refer to [extra modules to install]('pip install"
                " forte['audio_ext']) or 'pip install forte"
                ". Note that additional steps might apply to Linux"
                " users (refer to "
                "https://pysoundfile.readthedocs.io/en/latest/#installation)."
            ) from e
        self.soundfile = soundfile

    def _collect(self, audio_directory) -> Iterator[Any]:  # type: ignore
        r"""Should be called with param ``audio_directory`` which is a path to a
        folder containing audio files.

        Args:
            audio_directory: audio directory containing the files.

        Returns: Iterator over paths to audio files
        """
        return dataset_path_iterator(audio_directory, self.configs.file_ext)

    def _cache_key_function(self, audio_file: str) -> str:
        return os.path.basename(audio_file)

    def _parse_pack(self, file_path: str) -> Iterator[DataPack]:
        pack: DataPack = DataPack()

        # Read in audio data and store in DataPack
        audio, sample_rate = self.soundfile.read(
            file=file_path, **(self.configs.read_kwargs or {})
        )
        pack.set_audio(audio=audio, sample_rate=sample_rate)
        pack.pack_name = file_path

        yield pack

    @classmethod
    def default_configs(cls):
        r"""This defines a basic configuration structure for audio reader.

        Here:

          - file_ext (str): The file extension to find the target audio files
             under a specific directory path. Default value is ".flac".

          - read_kwargs (dict): A dictionary containing all the keyword
             arguments for `soundfile.read` method. For details, refer to
             https://pysoundfile.readthedocs.io/en/latest/#soundfile.read.
             Default value is None.

        Returns: The default configuration of audio reader.
        """
        return {"file_ext": ".flac", "read_kwargs": None}
